# Form implementation generated from reading ui file 'fitsettings.ui'
#
# Created by: PyQt6 UI code generator 6.7.0
#
# WARNING: Any manual changes made to this file will be lost when pyuic6 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt6 import QtCore, QtGui, QtWidgets


class Ui_fitparbox(object):
    def setupUi(self, fitparbox):
        fitparbox.setObjectName("fitparbox")
        fitparbox.resize(582, 447)
        self.gridLayout_2 = QtWidgets.QGridLayout(fitparbox)
        self.gridLayout_2.setObjectName("gridLayout_2")
        self.abundcutlabel = QtWidgets.QLabel(parent=fitparbox)
        self.abundcutlabel.setObjectName("abundcutlabel")
        self.gridLayout_2.addWidget(self.abundcutlabel, 6, 0, 1, 1)
        self.line = QtWidgets.QFrame(parent=fitparbox)
        self.line.setFrameShape(QtWidgets.QFrame.Shape.HLine)
        self.line.setFrameShadow(QtWidgets.QFrame.Shadow.Sunken)
        self.line.setObjectName("line")
        self.gridLayout_2.addWidget(self.line, 2, 0, 1, 2)
        self.wavecutlabel = QtWidgets.QLabel(parent=fitparbox)
        self.wavecutlabel.setObjectName("wavecutlabel")
        self.gridLayout_2.addWidget(self.wavecutlabel, 4, 0, 1, 1)
        self.convovcutlabel = QtWidgets.QLabel(parent=fitparbox)
        self.convovcutlabel.setObjectName("convovcutlabel")
        self.gridLayout_2.addWidget(self.convovcutlabel, 5, 0, 1, 1)
        self.abunditerlabel = QtWidgets.QLabel(parent=fitparbox)
        self.abunditerlabel.setObjectName("abunditerlabel")
        self.gridLayout_2.addWidget(self.abunditerlabel, 12, 0, 1, 1)
        self.wavecutvalue = QtWidgets.QDoubleSpinBox(parent=fitparbox)
        self.wavecutvalue.setMaximum(999.99)
        self.wavecutvalue.setSingleStep(0.01)
        self.wavecutvalue.setObjectName("wavecutvalue")
        self.gridLayout_2.addWidget(self.wavecutvalue, 4, 1, 1, 1)
        self.repfitvalue = QtWidgets.QSpinBox(parent=fitparbox)
        self.repfitvalue.setMaximum(10)
        self.repfitvalue.setObjectName("repfitvalue")
        self.gridLayout_2.addWidget(self.repfitvalue, 1, 1, 1, 1)
        self.contitervalue = QtWidgets.QSpinBox(parent=fitparbox)
        self.contitervalue.setMaximum(10000)
        self.contitervalue.setObjectName("contitervalue")
        self.gridLayout_2.addWidget(self.contitervalue, 10, 1, 1, 1)
        self.contiterlabel = QtWidgets.QLabel(parent=fitparbox)
        self.contiterlabel.setObjectName("contiterlabel")
        self.gridLayout_2.addWidget(self.contiterlabel, 10, 0, 1, 1)
        self.plotcutlabel = QtWidgets.QLabel(parent=fitparbox)
        self.plotcutlabel.setObjectName("plotcutlabel")
        self.gridLayout_2.addWidget(self.plotcutlabel, 7, 0, 1, 1)
        self.line_2 = QtWidgets.QFrame(parent=fitparbox)
        self.line_2.setFrameShape(QtWidgets.QFrame.Shape.HLine)
        self.line_2.setFrameShadow(QtWidgets.QFrame.Shadow.Sunken)
        self.line_2.setObjectName("line_2")
        self.gridLayout_2.addWidget(self.line_2, 8, 0, 1, 2)
        self.waveconviterlabel = QtWidgets.QLabel(parent=fitparbox)
        self.waveconviterlabel.setObjectName("waveconviterlabel")
        self.gridLayout_2.addWidget(self.waveconviterlabel, 11, 0, 1, 1)
        self.abundcutvalue = QtWidgets.QDoubleSpinBox(parent=fitparbox)
        self.abundcutvalue.setMaximum(999.99)
        self.abundcutvalue.setSingleStep(0.01)
        self.abundcutvalue.setObjectName("abundcutvalue")
        self.gridLayout_2.addWidget(self.abundcutvalue, 6, 1, 1, 1)
        self.cutvaltitle = QtWidgets.QLabel(parent=fitparbox)
        self.cutvaltitle.setAlignment(QtCore.Qt.AlignmentFlag.AlignCenter)
        self.cutvaltitle.setObjectName("cutvaltitle")
        self.gridLayout_2.addWidget(self.cutvaltitle, 3, 0, 1, 2)
        self.abunditervalue = QtWidgets.QSpinBox(parent=fitparbox)
        self.abunditervalue.setMaximum(10000)
        self.abunditervalue.setObjectName("abunditervalue")
        self.gridLayout_2.addWidget(self.abunditervalue, 12, 1, 1, 1)
        self.plotcutvalue = QtWidgets.QDoubleSpinBox(parent=fitparbox)
        self.plotcutvalue.setMaximum(999.99)
        self.plotcutvalue.setSingleStep(0.01)
        self.plotcutvalue.setObjectName("plotcutvalue")
        self.gridLayout_2.addWidget(self.plotcutvalue, 7, 1, 1, 1)
        self.fitmaxiterlabel = QtWidgets.QLabel(parent=fitparbox)
        self.fitmaxiterlabel.setAlignment(QtCore.Qt.AlignmentFlag.AlignCenter)
        self.fitmaxiterlabel.setObjectName("fitmaxiterlabel")
        self.gridLayout_2.addWidget(self.fitmaxiterlabel, 9, 0, 1, 2)
        self.convovcutvalue = QtWidgets.QDoubleSpinBox(parent=fitparbox)
        self.convovcutvalue.setMaximum(999.99)
        self.convovcutvalue.setSingleStep(0.01)
        self.convovcutvalue.setObjectName("convovcutvalue")
        self.gridLayout_2.addWidget(self.convovcutvalue, 5, 1, 1, 1)
        self.waveconvitervalue = QtWidgets.QSpinBox(parent=fitparbox)
        self.waveconvitervalue.setMaximum(10000)
        self.waveconvitervalue.setObjectName("waveconvitervalue")
        self.gridLayout_2.addWidget(self.waveconvitervalue, 11, 1, 1, 1)
        self.repfitlabel = QtWidgets.QLabel(parent=fitparbox)
        self.repfitlabel.setAlignment(QtCore.Qt.AlignmentFlag.AlignCenter)
        self.repfitlabel.setObjectName("repfitlabel")
        self.gridLayout_2.addWidget(self.repfitlabel, 0, 0, 1, 2)
        self.waveboundlabel = QtWidgets.QLabel(parent=fitparbox)
        self.waveboundlabel.setAlignment(QtCore.Qt.AlignmentFlag.AlignCenter)
        self.waveboundlabel.setObjectName("waveboundlabel")
        self.gridLayout_2.addWidget(self.waveboundlabel, 0, 3, 1, 2)
        self.waveboundmaxshiftlabel = QtWidgets.QLabel(parent=fitparbox)
        self.waveboundmaxshiftlabel.setObjectName("waveboundmaxshiftlabel")
        self.gridLayout_2.addWidget(self.waveboundmaxshiftlabel, 1, 3, 1, 1)
        self.waveboundmaxshiftvalue = QtWidgets.QDoubleSpinBox(parent=fitparbox)
        self.waveboundmaxshiftvalue.setObjectName("waveboundmaxshiftvalue")
        self.gridLayout_2.addWidget(self.waveboundmaxshiftvalue, 1, 4, 1, 1)
        self.line_3 = QtWidgets.QFrame(parent=fitparbox)
        self.line_3.setFrameShape(QtWidgets.QFrame.Shape.HLine)
        self.line_3.setFrameShadow(QtWidgets.QFrame.Shadow.Sunken)
        self.line_3.setObjectName("line_3")
        self.gridLayout_2.addWidget(self.line_3, 2, 3, 1, 2)
        self.convboundlabel = QtWidgets.QLabel(parent=fitparbox)
        self.convboundlabel.setAlignment(QtCore.Qt.AlignmentFlag.AlignCenter)
        self.convboundlabel.setObjectName("convboundlabel")
        self.gridLayout_2.addWidget(self.convboundlabel, 3, 3, 1, 2)
        self.convboundminlabel = QtWidgets.QLabel(parent=fitparbox)
        self.convboundminlabel.setObjectName("convboundminlabel")
        self.gridLayout_2.addWidget(self.convboundminlabel, 4, 3, 1, 1)
        self.convboundminvalue = QtWidgets.QDoubleSpinBox(parent=fitparbox)
        self.convboundminvalue.setObjectName("convboundminvalue")
        self.gridLayout_2.addWidget(self.convboundminvalue, 4, 4, 1, 1)
        self.convboundmaxlabel = QtWidgets.QLabel(parent=fitparbox)
        self.convboundmaxlabel.setObjectName("convboundmaxlabel")
        self.gridLayout_2.addWidget(self.convboundmaxlabel, 5, 3, 1, 1)
        self.convboundmaxvalue = QtWidgets.QDoubleSpinBox(parent=fitparbox)
        self.convboundmaxvalue.setObjectName("convboundmaxvalue")
        self.gridLayout_2.addWidget(self.convboundmaxvalue, 5, 4, 1, 1)
        self.repfitvaluelabel = QtWidgets.QLabel(parent=fitparbox)
        self.repfitvaluelabel.setObjectName("repfitvaluelabel")
        self.gridLayout_2.addWidget(self.repfitvaluelabel, 1, 0, 1, 1)
        self.line_4 = QtWidgets.QFrame(parent=fitparbox)
        self.line_4.setFrameShape(QtWidgets.QFrame.Shape.HLine)
        self.line_4.setFrameShadow(QtWidgets.QFrame.Shadow.Sunken)
        self.line_4.setObjectName("line_4")
        self.gridLayout_2.addWidget(self.line_4, 6, 3, 1, 2)
        self.contfitparamlabel = QtWidgets.QLabel(parent=fitparbox)
        self.contfitparamlabel.setAlignment(QtCore.Qt.AlignmentFlag.AlignCenter)
        self.contfitparamlabel.setObjectName("contfitparamlabel")
        self.gridLayout_2.addWidget(self.contfitparamlabel, 7, 3, 1, 2)
        self.contfitparalphalabel = QtWidgets.QLabel(parent=fitparbox)
        self.contfitparalphalabel.setObjectName("contfitparalphalabel")
        self.gridLayout_2.addWidget(self.contfitparalphalabel, 8, 3, 1, 1)
        self.contfitparalphavalue = QtWidgets.QDoubleSpinBox(parent=fitparbox)
        self.contfitparalphavalue.setObjectName("contfitparalphavalue")
        self.gridLayout_2.addWidget(self.contfitparalphavalue, 8, 4, 1, 1)
        self.contfitparepslabel = QtWidgets.QLabel(parent=fitparbox)
        self.contfitparepslabel.setObjectName("contfitparepslabel")
        self.gridLayout_2.addWidget(self.contfitparepslabel, 9, 3, 1, 1)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.label = QtWidgets.QLabel(parent=fitparbox)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Fixed, QtWidgets.QSizePolicy.Policy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label.sizePolicy().hasHeightForWidth())
        self.label.setSizePolicy(sizePolicy)
        font = QtGui.QFont()
        font.setPointSize(12)
        self.label.setFont(font)
        self.label.setObjectName("label")
        self.horizontalLayout.addWidget(self.label)
        self.contfitparepsvalue = QtWidgets.QSpinBox(parent=fitparbox)
        self.contfitparepsvalue.setObjectName("contfitparepsvalue")
        self.horizontalLayout.addWidget(self.contfitparepsvalue)
        self.label_2 = QtWidgets.QLabel(parent=fitparbox)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Fixed, QtWidgets.QSizePolicy.Policy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label_2.sizePolicy().hasHeightForWidth())
        self.label_2.setSizePolicy(sizePolicy)
        font = QtGui.QFont()
        font.setPointSize(12)
        self.label_2.setFont(font)
        self.label_2.setObjectName("label_2")
        self.horizontalLayout.addWidget(self.label_2)
        self.gridLayout_2.addLayout(self.horizontalLayout, 9, 4, 1, 1)
        self.line_5 = QtWidgets.QFrame(parent=fitparbox)
        self.line_5.setFrameShape(QtWidgets.QFrame.Shape.VLine)
        self.line_5.setFrameShadow(QtWidgets.QFrame.Shadow.Sunken)
        self.line_5.setObjectName("line_5")
        self.gridLayout_2.addWidget(self.line_5, 0, 2, 13, 1)
        self.okcancelbutton = QtWidgets.QDialogButtonBox(parent=fitparbox)
        self.okcancelbutton.setOrientation(QtCore.Qt.Orientation.Horizontal)
        self.okcancelbutton.setStandardButtons(QtWidgets.QDialogButtonBox.StandardButton.Cancel|QtWidgets.QDialogButtonBox.StandardButton.Ok)
        self.okcancelbutton.setObjectName("okcancelbutton")
        self.gridLayout_2.addWidget(self.okcancelbutton, 12, 3, 1, 2)

        self.retranslateUi(fitparbox)
        self.okcancelbutton.accepted.connect(fitparbox.accept) # type: ignore
        self.okcancelbutton.rejected.connect(fitparbox.reject) # type: ignore
        QtCore.QMetaObject.connectSlotsByName(fitparbox)

    def retranslateUi(self, fitparbox):
        _translate = QtCore.QCoreApplication.translate
        fitparbox.setWindowTitle(_translate("fitparbox", "Fit Settings"))
        self.abundcutlabel.setText(_translate("fitparbox", "Abundance"))
        self.wavecutlabel.setToolTip(_translate("fitparbox", "Continuum range is not used in \"Eq. Width\" method, but Wave. Shift is."))
        self.wavecutlabel.setText(_translate("fitparbox", "Cont/Wave. Shift"))
        self.convovcutlabel.setToolTip(_translate("fitparbox", "Can not be higher than the value of Continuum/Wave. Shift."))
        self.convovcutlabel.setText(_translate("fitparbox", "Convolution"))
        self.abunditerlabel.setText(_translate("fitparbox", "Abundance"))
        self.wavecutvalue.setToolTip(_translate("fitparbox", "Continuum range is not used in \"Eq. Width\" method, but Wave. Shift is."))
        self.contiterlabel.setText(_translate("fitparbox", "Continuum"))
        self.plotcutlabel.setText(_translate("fitparbox", "Plot"))
        self.waveconviterlabel.setText(_translate("fitparbox", "Wave. Shift/Conv."))
        self.cutvaltitle.setToolTip(_translate("fitparbox", "The range that the fit will be applied. The value represents the full range (half of the value left to the fitted line and half to the right)."))
        self.cutvaltitle.setText(_translate("fitparbox", "Interaction range (Angstrom) for each fit"))
        self.fitmaxiterlabel.setText(_translate("fitparbox", "Fit Maximum Iterations"))
        self.convovcutvalue.setToolTip(_translate("fitparbox", "Can not be higher than the value of Continuum/Wave. Shift."))
        self.repfitlabel.setToolTip(_translate("fitparbox", "The number of times that it will try to fit each line. The subsequent iteration uses the previous result as start point. Default is 2."))
        self.repfitlabel.setText(_translate("fitparbox", "Forced Loop Iterations"))
        self.waveboundlabel.setText(_translate("fitparbox", "Wave. Shift (Angstrom) Boundaries"))
        self.waveboundmaxshiftlabel.setText(_translate("fitparbox", "Max. Shift (each side)"))
        self.convboundlabel.setText(_translate("fitparbox", "Convolution Boundaries"))
        self.convboundminlabel.setText(_translate("fitparbox", "Minimum"))
        self.convboundmaxlabel.setText(_translate("fitparbox", "Maximum"))
        self.repfitvaluelabel.setText(_translate("fitparbox", "Value"))
        self.contfitparamlabel.setText(_translate("fitparbox", "Continuum Fit Parameters"))
        self.contfitparalphalabel.setText(_translate("fitparbox", "Tolerance (alpha)"))
        self.contfitparepslabel.setText(_translate("fitparbox", "Error (epsilon)"))
        self.label.setText(_translate("fitparbox", "10 ^ ( -"))
        self.label_2.setText(_translate("fitparbox", ")"))


if __name__ == "__main__":
    import sys
    app = QtWidgets.QApplication(sys.argv)
    fitparbox = QtWidgets.QDialog()
    ui = Ui_fitparbox()
    ui.setupUi(fitparbox)
    fitparbox.show()
    sys.exit(app.exec())



"""
| MEAFS Voigt Functions
| Matheus J. Castro

| Voigt, Gaussian, Lorentzian module functions.
"""

from scipy.optimize import minimize
import pandas as pd
import numpy as np

from . import fit_functions as ff


def gaussian(x, b, c, a=1, d=0):
    """
    Gaussian function.

    .. math:: f(x) = a \\cdot \\exp\\left[\\frac{-(x - b)^2}{c \\cdot \\sqrt{2 \\cdot \\pi}}\\right] + d
        :label: gauss


    :param x: can be a list or a number.
    :param b: a single number.
    :param c: a single number.
    :param a: a single number.
    :param d: a single number.
    :return: :math:`f(x)` as a number or a list.
    """

    return a * np.exp(-(x - b)**2 / (c * np.sqrt(2 * np.pi))) + d


def lorentzian(x, b, c, a=1, d=0):
    """
    Lorentzian function.

    .. math:: g(x) = a \\cdot \\exp\\left[\\frac{c}{\\pi \\cdot ((x - b)^2 + c^2)}\\right] + d
        :label: loren

    :param x: can be a list or a number.
    :param b: a single number.
    :param c: a single number.
    :param a: a single number.
    :param d: a single number.
    :return: :math:`g(x)` as a number or a list.
    """

    return a * c / (np.pi * ((x - b)**2 + c**2)) + d


def voigt(x, b, c, a, d):
    """
    Voigt function.

    .. math::
        h(x) = a \\cdot f(x) \\cdot g(x) + d

    With :math:`f(x)` being the :eq:`gauss` and :math:`g(x)` being the :eq:`loren`.

    :param x: can be a list or a number.
    :param b: a single number.
    :param c: a single number.
    :param a: a single number.
    :param d: a single number.
    :return: :math:`h(x)` as a number or a list.
    """

    return a * gaussian(x, b, c) * lorentzian(x, b, c) + d


def find_func(type_func):
    """
    Determines the function to be called: Gaussian, Lorentzian or Voigt.

    :param type_func: string with the name of the function.
    :return: the function itself.
    """

    if type_func == "Gaussian":
        func = gaussian
    elif type_func == "Lorentzian":
        func = lorentzian
    elif type_func == "Voigt":
        func = voigt
    else:
        print("Function not recognized.")
        func = None
    return func


def optimize_spec(spec_obs_cut, type_synth, lamb, continuum, convovbound=None,
                  wavebound=None, iterac=100):
    """
    Fit of the Convolution and the Wavelength Shift using the minimization of the :math:`\\chi^2`
    with the Nelder-Mead method.

    :param spec_obs_cut: spectrum data.
    :param type_synth: type of the function to apply.
    :param lamb: current wavelength.
    :param continuum: continuum value.
    :param convovbound: range to fit the convolution.
    :param wavebound: range to fit the wavelength shift.
    :param iterac: maximum allowed iterations of the Nelder-Mead method.
    :return: the optimized parameters, the value of the minimum :math:`\\chi^2` and the spectrum
             generated with the best parameters.
    """

    if convovbound is None:
        convovbound = [0, 1]
    if wavebound is None:
        wavebound = [lamb-1, lamb+1]
    else:
        wavebound = [lamb-wavebound, lamb+wavebound]

    func = find_func(type_synth[1])

    x = np.linspace(min(spec_obs_cut.iloc[:, 0]), max(spec_obs_cut.iloc[:, 0]), 1000)

    a = -(continuum - spec_obs_cut.iloc[ff.bisec(spec_obs_cut, lamb), 1])
    b = lamb
    c = type_synth[2]
    d = continuum

    def fit(guess):
        bfit = guess[0]
        cfit = guess[1]
        sp_fit = [x, func(x, bfit, cfit, a, d)]
        return ff.chi2(spec_obs_cut, sp_fit)

    b, c = minimize(fit, np.array([b, c]), method='Nelder-Mead', bounds=[wavebound, convovbound],
                    options={"maxiter": iterac}).x

    # opt_pars = [lamb_desloc, continuum, convol]
    opt_pars = [b - lamb, d, c]
    spec_fit = pd.DataFrame({0: x, 1: func(x, b, c, a, d)})
    chi = ff.chi2(spec_obs_cut, spec_fit)

    return opt_pars, chi, spec_fit


def optimize_abund(spec_obs_cut, type_synth, lamb, opt_pars, iterac=100):
    """
    Fit of the Abundance using the minimization of the :math:`\\chi^2`
    with the Nelder-Mead method.

    :param spec_obs_cut: spectrum data.
    :param type_synth: type of the function to apply.
    :param lamb: current wavelength.
    :param opt_pars: the Continuum, Convolution and Wavelength Shift parameters.
    :param iterac: maximum allowed iterations of the Nelder-Mead method.
    :return: the abundance, the value of the minimum :math:`\\chi^2` and the spectrum
             generated with the best parameters.
    """

    func = find_func(type_synth[1])

    x = np.linspace(min(spec_obs_cut.iloc[:, 0]), max(spec_obs_cut.iloc[:, 0]), 1000)

    a = - (opt_pars[1] - spec_obs_cut.iloc[ff.bisec(spec_obs_cut, lamb), 1])
    b = lamb
    c = opt_pars[2]
    d = opt_pars[1]

    def fit(guess):
        afit = guess[0]
        sp_fit = [x, func(x, b, c, afit, d)]
        return ff.chi2(spec_obs_cut, sp_fit)

    a = minimize(fit, np.array([a]), method='Nelder-Mead',
                 options={"maxiter": iterac}).x

    par = a
    spec_fit = pd.DataFrame({0: x, 1: func(x, b, c, a, d)})
    chi = ff.chi2(spec_obs_cut, spec_fit)

    return par, chi, spec_fit



"""
| MEAFS TurboEspectrum Functions
| Matheus J. Castro

| TurboSpectrum module functions.
"""

from scipy.optimize import minimize
import pandas as pd
import numpy as np
import subprocess

from . import fit_functions as ff


def check_elem_configfl(fl_name, elem):
    """
    Find the desired element in the Turbospectrum2019 configuration file.

    :param fl_name: file name of the TurboSpectrum configuration file.
    :param elem: element to look for.
    :return: true if finds it; false if not.
    """

    with open(fl_name, 'r') as file:
        fl = file.read()

    elem = "foreach " + elem + "_ab"
    pos = fl.find(elem)

    if pos == -1:
        return False
    else:
        return True


def change_spec_range_configfl(fl_name, lamb, cut_val):
    """
    Change the spectrum range in Turbospectrum2019 configuration file

    :param fl_name: file name of the TurboSpectrum configuration file.
    :param lamb: central wavelength.
    :param cut_val: range to be applied.
    """

    spec_range = [lamb - cut_val, lamb + cut_val]

    with open(fl_name, 'r') as file:
        fl = file.read()

    for i, str_i in enumerate(["set lam_min    = ", "set lam_max    = "]):
        pos = fl.find(str_i)
        # noinspection PyListCreation
        pos = [pos + fl[pos:].find("'") + 1]
        # noinspection
        pos.append(pos[0] + fl[pos[0]:].find("'"))
        fl = fl[:pos[0]] + str(spec_range[i]) + fl[pos[1]:]

    with open(fl_name, "w") as file:
        file.write(fl)


def change_abund_configfl(fl_name, elem, abund=None, find=True):
    """
    Change the element abundance in Turbospectrum2019 configuration file
    or find the current value.

    :param fl_name: file name of the TurboSpectrum configuration file.
    :param elem: element to be changed.
    :param abund: desired abundance.
    :param find: true if it is only to get the current value.
    :return: abundance.
    """

    with open(fl_name, 'r') as file:
        fl = file.read()

    elem = "foreach " + elem + "_ab"
    pos = fl.find(elem)
    pos += fl[pos:].find("(") + 1
    pos_end = pos + fl[pos:].find(")")

    if find:
        abund = float(fl[pos:pos_end])
    else:
        fl = fl[:pos] + str(abund) + fl[pos_end:]

        with open(fl_name, "w") as file:
            file.write(fl)

    return abund


def run_configfl(config_fl):
    """
    Run Turbospectrum2019.

    :param config_fl: file name of the TurboSpectrum configuration file.
    """

    run = config_fl.split("/")[-1]
    config_fodler = config_fl.removesuffix(run)

    subprocess.check_output("cd \"{}\" && ./{}".format(config_fodler, run), shell=True)


def optimize_spec(spec_obs_cut, spec_conv, lamb, cut_val, continuum, init=None, iterac=100, convovbound=None,
                  wavebound=None):
    """
    Fit of the Convolution, Wavelength Shift and Continuum using the minimization of the :math:`\\chi^2`
    with the Nelder-Mead method.

    :param spec_obs_cut: spectrum data.
    :param spec_conv: synthetic spectrum data.
    :param lamb: current wavelength.
    :param cut_val: range to cut the spectrum for the convolution fit.
    :param continuum: continuum value.
    :param init: initial guess values for the Wavelength Shift, Continuum and Convolution.
    :param iterac: maximum allowed iterations of the Nelder-Mead method.
    :param convovbound: range to fit the convolution.
    :param wavebound: range to fit the wavelength shift.
    :return: the optimized parameters
    """

    # Define initial values if not given
    if init is None:
        init = [0, 1, 3.85]
    if convovbound is None:
        convovbound = [3.5, 4.2]
    if wavebound is None:
        wavebound = [lamb-1, lamb+1]
    else:
        wavebound = [-wavebound, +wavebound]

    # First, apply a convolution of 3.85
    sp_convoluted = ff.spec_operations(spec_conv.copy(), convol=init[2])

    # Find the best parameters for lambda shift
    def opt_desloc_continuum(guess):
        sp_fit = ff.spec_operations(sp_convoluted.copy(), lamb_desloc=guess[0], continuum=guess[1], convol=0)
        return ff.chi2(spec_obs_cut, sp_fit)
    pars = minimize(opt_desloc_continuum, np.array([init[0], continuum]), method='Nelder-Mead',
                    options={"maxiter": iterac}, bounds=[wavebound, [continuum*0.9, continuum*1.1]]).x

    # pars = [pars[0], continuum]

    spec_obs_cut = ff.cut_spec(spec_obs_cut, lamb, cut_val=cut_val[1])
    spec_conv = ff.cut_spec(spec_conv, lamb, cut_val=cut_val[1])

    # Finally, find the best convolution
    def opt_convolution(guess):
        # noinspection PyTypeChecker
        sp_fit = ff.spec_operations(spec_conv.copy(), lamb_desloc=pars[0], continuum=pars[1], convol=guess[0])
        return ff.chi2(spec_obs_cut, sp_fit)
    par = minimize(opt_convolution, np.array([init[2]]), method='Nelder-Mead', options={"maxiter": iterac},
                   bounds=[convovbound]).x

    pars = np.append(pars, par, axis=0)
    # pars = [0, 1, 1]

    return pars


def optimize_abund(spec_obs_cut, config_fl, conv_name, elem, opt_pars, par, abund_lim, iterac=10):
    """
    Fit of the Abundance using the minimization of the :math:`\\chi^2`
    with the Nelder-Mead method.

    :param spec_obs_cut: spectrum data.
    :param config_fl: file name of the TurboSpectrum configuration file.
    :param conv_name: file name of the TurboSpectrum output file.
    :param elem: element to be fitted.
    :param opt_pars: the Continuum, Convolution and Wavelength Shift parameters.
    :param par: initial guess for the optimization.
    :param abund_lim: range to try the fit.
    :param iterac: maximum allowed iterations of the Nelder-Mead method.
    :return: the abundance, the value of the minimum :math:`\\chi^2` and the spectrum
             generated with the best parameters.
    """

    def opt_abund(abund):
        change_abund_configfl(config_fl, elem, find=False, abund=abund[0])
        run_configfl(config_fl)
        spec = pd.read_csv(conv_name, header=None, delimiter=r"\s+")
        # noinspection PyTypeChecker
        spec = ff.spec_operations(spec.copy(), lamb_desloc=opt_pars[0], continuum=opt_pars[1],
                                  convol=opt_pars[2])

        chi = ff.chi2(spec_obs_cut, spec)
        return chi

    par = minimize(opt_abund, np.array(par), method='Nelder-Mead', options={"maxiter": iterac},
                   bounds=[[par[0] - abund_lim, par[0] + abund_lim]]).x

    spec_conv = pd.read_csv(conv_name, header=None, delimiter=r"\s+")
    spec_fit = ff.spec_operations(spec_conv, lamb_desloc=opt_pars[0], continuum=opt_pars[1],
                                  convol=opt_pars[2])
    chi = ff.chi2(spec_obs_cut, spec_fit)

    return par, chi, spec_fit



#!/usr/bin/env python3
"""
| MEAFS Abundance Fit
| Matheus J. Castro

| Main fit code to do all the operations with the spectrum, call the modules to generate
  the synthetic spectrum, plot the curves and more.
"""

from specutils.analysis import equivalent_width
from PyQt6 import QtWidgets, QtCore
from specutils import Spectrum1D
import matplotlib.pyplot as plt
import astropy.units as u
from pathlib import Path
import pandas as pd
import numpy as np
import time
import sys
import os

from . import fit_functions as ff
from . import voigt_functions as vf
from . import turbospec_functions as tf
from . import abundance_plot as ap

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "..", "..")))
# noinspection PyUnresolvedReferences
from meafs_code import __version__


def open_linelist_refer_fl(list_name):
    """
    Open the Linelist using PANDAS.

    :param list_name: file name of the linelist.
    :return: the pandas dataframe with the linelist.
    """

    return pd.read_csv(list_name, header=None, comment="#")


def open_spec_obs(observed_name, delimiter=None, increment=1):
    """
    Open the observed spectrum file using PANDAS.

    :param observed_name: file name with the spectrum.
    :param delimiter: delimiter type of the file.
    :param increment: increment to get the delimiter automatically
                      (previously necessary in the older no-gui version).
    :return: the pandas dataframe with the spectrum.
    """

    spec_obs = []

    for i in range(0, len(observed_name), increment):
        if increment == 2:
            delimiter = observed_name[i + 1]
        spec_obs.append(pd.read_csv(observed_name[i], header=None, delimiter=delimiter, comment="#"))
    return spec_obs


def open_previous(linelist, columns_names, fl_name=Path("found_values.csv")):
    """
    Open the previous results and analyse it.

    :param linelist: linelist object.
    :param columns_names: names of the columns in the file.
    :param fl_name: file name.
    :return: the new linelist and the pandas dataframe with the previous results.
    """

    try:
        # Try to open
        prev = pd.read_csv(fl_name, delimiter=",", dtype=str)
    except (FileNotFoundError, pd.errors.EmptyDataError):
        # If file not found or empty database, create a new one and return
        return linelist, pd.DataFrame(columns=columns_names)

    for i in prev.columns:
        if i != "Element" and i != "Chi" and i != "Equiv Width Obs (A)" and i != "Equiv Width Fit (A)":
            prev[i] = prev[i].astype(float)

    # Reads the existing data
    new_linelist = pd.DataFrame(columns=[0, 1])
    for i in range(len(linelist)):
        elem = linelist.iloc[i][0]
        lamb = linelist.iloc[i][1]
        ind = prev[(prev["Element"] == elem) & (prev["Lambda (A)"] == float(lamb))].index
        if ind.tolist() == []:
            new_linelist.loc[i] = linelist.iloc[i]

    return new_linelist, prev


def plot_spec_gui(specs, canvas, ax, plot_line_refer):
    """
    Plot the spectrum in the GUI.

    :param specs: spectrum data to be plotted.
    :param canvas: canvas to plot.
    :param ax: ax to plot.
    :param plot_line_refer: array to save the reference label of the plots.
    :return: the actualized ``plot_line_refer`` array.
    """

    if specs != []:
        min_all_old = min(specs[0].iloc[:, 0])
        max_all_old = max(specs[0].iloc[:, 0])
    else:
        min_all_old = 0
        max_all_old = 1

    same_line_plots = plot_line_refer[plot_line_refer["elem"] == "spec"]
    for line in range(len(same_line_plots)):
        same_line_plots.iloc[line]["refer"].remove()
        ind = plot_line_refer[plot_line_refer["refer"] == same_line_plots.iloc[line]["refer"]].index
        plot_line_refer = plot_line_refer.drop(ind)
    plot_line_refer.reset_index(drop=True, inplace=True)

    for spec in specs:
        lineplot = ax.plot(spec.iloc[:, 0], spec.iloc[:, 1])
        min_all = min(spec.iloc[:, 0])
        max_all = max(spec.iloc[:, 0])
        if min_all < min_all_old:
            min_all_old = min_all
        if max_all > max_all_old:
            max_all_old = max_all

        plot_line_refer.loc[len(plot_line_refer)] = {"elem": "spec", "wave": "all", "refer": lineplot[0]}

    ax.set_xlim(min_all_old, max_all_old)
    canvas.draw()

    return plot_line_refer


def plot_spec(spec1, spec2, spec3, lamb, elem, folder, show=False, save=True):
    """
    Plot the spectra (previously necessary in the older no-gui version).

    :param spec1: first spectrum data.
    :param spec2: second spectrum data.
    :param spec3: third spectrum data.
    :param lamb: current wavelength.
    :param elem: current element.
    :param folder: directory to save the plot.
    :param show: true to show with the ``matplotlib.pyplot.show()``.
    :param save: true to save the spectrum.
    """


    plt.figure(figsize=(16, 9))

    plt.plot(spec1[0], spec1[1], label="Observed")
    plt.plot(spec2[0], spec2[1], label="Abundance Fit")
    if spec3 is not None:
        plt.plot(spec3[0], spec3[1], label="Final Fit")
    if lamb is not None:
        plt.axvline(lamb, color="red", label="Reference")

    plt.legend()

    if show:
        plt.show()
    if save:
        plt.savefig(Path(folder).joinpath("On_time_Plots", "fit_{}_{}_ang.pdf".format(elem, lamb)))
    plt.close()


def plot_spec_ui(spec_fit_arr, folder, elem, lamb, order, ax, canvas, plot_line_refer, vline=True):
    """
    Plot the current line results in the GUI.

    :param spec_fit_arr: spectrum line data.
    :param folder: directory to save.
    :param elem: current element.
    :param lamb: current wavelength.
    :param order: current order of the line.
    :param ax: ax to plot.
    :param canvas: canvas to plot.
    :param plot_line_refer: array to save the reference label of the plots.
    :param vline: if it also plots a vertical line or not.
    :return: the actualized ``plot_line_refer`` array.
    """

    # noinspection PySimplifyBooleanCheck
    if spec_fit_arr != []:
        min_all_old = min(spec_fit_arr[0].iloc[:, 0])
        max_all_old = max(spec_fit_arr[0].iloc[:, 0])
    else:
        min_all_old = 0
        max_all_old = 1

    for j, sp in enumerate(spec_fit_arr):
        ind = plot_line_refer[(plot_line_refer["elem"] == elem + order) &
                              (plot_line_refer["wave"] == lamb)].index
        for line in plot_line_refer.loc[ind, "refer"]: line.remove()
        plot_line_refer.drop(ind, inplace=True)
        plot_line_refer.reset_index(drop=True, inplace=True)

        lineplot = ax.plot(sp.iloc[:, 0], sp.iloc[:, 1], "--", linewidth=1.5)

        min_all = min(sp.iloc[:, 0])
        max_all = max(sp.iloc[:, 0])
        if min_all < min_all_old:
            min_all_old = min_all
        if max_all > max_all_old:
            max_all_old = max_all

        plot_line_refer.loc[len(plot_line_refer)] = {"elem": elem+order, "wave": lamb, "refer": lineplot[0]}

        if vline:
            axvlineplot = ax.axvline(lamb, ls="-.", c="red", linewidth=.5)
            plot_line_refer.loc[len(plot_line_refer)] = {"elem": elem+order, "wave": lamb, "refer": axvlineplot}

        sp.to_csv(Path(folder).joinpath("On_time_Plots",
                                        "fit_{}_{}_ang_{}.csv".format(elem + order, lamb, j + 1)),
                  index=False,
                  float_format="%.4f",
                  header=None)

    ax.set_xlim(min_all_old, max_all_old)
    canvas.draw()

    return plot_line_refer


def check_order(elem):
    """
    | Check if the current element has any type of order written in the end of the string.
    | Supported types are: roman numbers *(I, II, III...)* and western digits *(1, 2, 3...)*.

    :param elem:
    :return: the element without the order and the order itself.
    """

    if elem[1].isdigit() or elem[1] == "I" or elem[1] == "V" or elem[1] == "X" or elem[1] == " ":
        order = elem[1:]
        elem = elem[0]
    else:
        order = elem[2:]
        elem = elem[0:2]

    return elem, order


def plot_abund_nofit(elem, lamb, abundplot, refer_fl, folder, type_synth,
                     cut_val=None, canvas=None, ax=None, plot_line_refer=None,
                     opt_pars=None):
    """
    Plot in the GUI the specified parameters.

    :param elem: current element.
    :param lamb: current wavelength.
    :param abundplot: abundance to be plotted.
    :param refer_fl: reference abundance dataframe.
    :param folder: directory to save.
    :param type_synth: type of the current synthetics spectrum generator.
    :param cut_val: range to plot.
    :param canvas: canvas to plot.
    :param ax: ax to plot.
    :param plot_line_refer: array to save the reference label of the plots.
    :param opt_pars: parameters for the convolution, wavelength shift and continuum.
    :return: the actualized ``plot_line_refer`` array.
    """

    if type_synth[0] == "TurboSpectrum":
        conv_name = type_synth[1]
        config_fl = type_synth[2]
    else:
        conv_name = None
        config_fl = None

    lamb = float(lamb)

    elem, order = check_order(elem)

    # Try to find element abundance reference. If not found, 0 is used.
    try:
        abund_val_refer = float(refer_fl.loc[elem, "value"]) if not refer_fl.loc[elem].isnull().item() else 0
    except KeyError:
        abund_val_refer = 0

    if type_synth[0] == "Equivalent Width":
        func = vf.find_func(type_synth[1])
        x = np.linspace(lamb-cut_val[3], lamb+cut_val[3], 1000)
        spec_fit = pd.DataFrame({0: x, 1: func(x, opt_pars[0] + lamb, opt_pars[2], abundplot, opt_pars[1])})
    elif type_synth[0] == "TurboSpectrum":
        tf.change_abund_configfl(config_fl, elem, find=False, abund=abundplot)
        tf.change_spec_range_configfl(config_fl, lamb, cut_val[3])
        tf.run_configfl(config_fl)
        # Return the Turbospectrum2019 configuration file to original abundance value
        tf.change_abund_configfl(config_fl, elem, find=False, abund=abund_val_refer)

        spec_conv = pd.read_csv(conv_name, header=None, delimiter=r"\s+")

        # noinspection PyTypeChecker
        spec_fit = ff.spec_operations(spec_conv.copy(), lamb_desloc=opt_pars[0], continuum=opt_pars[1],
                                      convol=opt_pars[2])
    else:
        return plot_line_refer

    spec_fit_arr = [spec_fit]
    plot_line_refer = plot_spec_ui(spec_fit_arr, folder, elem, lamb, order, ax, canvas, plot_line_refer)

    return plot_line_refer


def fit_abundance(linelist, spec_obs, refer_fl, folder, type_synth, cut_val=None,
                  abund_lim_df=1., restart=False, save_name="found_values.csv",
                  ui=None, canvas=None, ax=None, plot_line_refer=None,
                  opt_pars=None, repfit=2, max_iter=None, convovbound=None,
                  contpars=None, wavebound=None, only_abund_ind=None):
    """
    Main function to analyse the spectrum and find the fit values for it

    :param linelist: linelist dataframe.
    :param spec_obs: spectrum data.
    :param refer_fl: reference abundace dataframe.
    :param folder: directory to save.
    :param type_synth: type of the current synthetics spectrum generator.
    :param cut_val: ranges to cut the spectrum.
    :param abund_lim_df: default value of the range of the allowed abundance.
    :param restart: if it should ignore past results in the database.
    :param save_name: file name of the previous results.
    :param ui: main GUI QT object.
    :param canvas: canvas to plot.
    :param ax: ax to plot.
    :param plot_line_refer: array to save the reference label of the plots.
    :param opt_pars: convolution, wavelength shift and continuum initial guess.
    :param repfit: number of iterations of the main fit function.
    :param max_iter: maximum allowed iterations of the Nelder-Mead method.
    :param convovbound: range to fit the convolution.
    :param contpars: the calibration values of the overall continuum fit method.
    :param wavebound: range to fit the wavelength shift.
    :param only_abund_ind: change only abundance without fitting other parameters at this index.
    :return: dataframe with the results of the fit, the actualized
             ``ax`` array and the actualized ``plot_line_refer`` array.
    """

    if type_synth[0] == "TurboSpectrum":
        conv_name = type_synth[1]
        config_fl = type_synth[2]
    else:
        conv_name = None
        config_fl = None

    # For each fit parameter, a spectrum range can be selected
    # a value of 5, will select a total of 10 Angstroms with the element line wavelength in the middle
    if cut_val is None:
        # spec range vals for [continuum, convolution, abundance, plot]
        cut_val = [10/2, 3/2, .4/2, 1/2]

    columns_names = ["Element", "Lambda (A)", "Lamb Shift", "Continuum", "Convolution",
                     "Refer Abundance", "Fit Abundance", "Differ", "Chi", "Equiv Width Obs (A)",
                     "Equiv Width Fit (A)"]

    if restart:
        found_val = pd.DataFrame(columns=columns_names)
    else:
        linelist, found_val = open_previous(linelist, columns_names, fl_name=Path(folder).joinpath(save_name))

        if only_abund_ind is not None:
            linelist = pd.DataFrame(columns=[0, 1])
            linelist.loc[0] = [found_val.loc[only_abund_ind, "Element"],
                               found_val.loc[only_abund_ind, "Lambda (A)"]]

        if ui is not None:
            rowpos = ui.abundancetable.rowCount()
            for i in range(len(found_val)):
                elem = found_val.iloc[i, 0]
                lamb = float(found_val.iloc[i, 1])

                elem, order = check_order(elem)

                ui.abundancetable.insertRow(rowpos+i)
                ui.abundancetable.setItem(rowpos+i, 0, QtWidgets.QTableWidgetItem(str(elem)+str(order)))
                ui.abundancetable.setItem(rowpos+i, 1, QtWidgets.QTableWidgetItem(str(lamb)))
                path = Path(folder).joinpath("On_time_Plots")
                count = 0
                while True:
                    file = path.joinpath("fit_{}_{}_ang_{}.csv".format(elem+order, lamb, count + 1))

                    if os.path.isfile(file):
                        data = pd.read_csv(file)
                        count += 1

                        ind = plot_line_refer[(plot_line_refer["elem"] == elem + order) &
                                              (plot_line_refer["wave"] == lamb)].index
                        for line in plot_line_refer.loc[ind, "refer"]: line.remove()
                        plot_line_refer.drop(ind, inplace=True)
                        plot_line_refer.reset_index(drop=True, inplace=True)

                        lineplot = ax.plot(data.iloc[:, 0], data.iloc[:, 1], "--", linewidth=1.5)
                        axvlineplot = ax.axvline(lamb, ls="-.", c="red", linewidth=.5)
                        plot_line_refer.loc[len(plot_line_refer)] = {"elem": elem+order, "wave": lamb,
                                                                     "refer": lineplot[0]}
                        plot_line_refer.loc[len(plot_line_refer)] = {"elem": elem+order, "wave": lamb,
                                                                     "refer": axvlineplot}
                    else:
                        break
                canvas.draw()

    if opt_pars is None:
        continuum, cont_err = ff.fit_continuum(spec_obs, contpars=contpars, iterac=max_iter[0])
    else:
        continuum = None

    # For each line in the linelist file
    for i in range(len(linelist)):
        elem = linelist.iloc[i][0]
        lamb = float(linelist.iloc[i][1])

        if ui is not None:
            ui.linedefvalue.setText("Element {}, line {}".format(elem, lamb))
            # Allow QT to actualize the UI while in the loop
            QtCore.QCoreApplication.processEvents()
            if ui.stop_state:
                ui.stop_state = False
                return found_val, ax, plot_line_refer

        print("Analysing the element {} for lambda {}. Line {} of {}.".format(elem, lamb, i+1, len(linelist)))

        elem, order = check_order(elem)

        # Try to find element abundance reference. If not found, 0 is used.
        try:
            abund_val_refer = float(refer_fl.loc[elem, "value"]) if not refer_fl.loc[elem].isnull().item() else 0
        except KeyError:
            abund_val_refer = 0

        abund_lim = abund_lim_df if abund_val_refer != 0 else 3

        # Check whether the element exists in turbospectrum config file
        if type_synth[0] == "TurboSpectrum" and not tf.check_elem_configfl(config_fl, elem):
            print("Element not in TurboSpectrum Configuration file")
            continue
        # If the lamb is not between the range of the spec, skip the iteration
        if not spec_obs.iloc[0][0] <= lamb <= spec_obs.iloc[-1][0]:
            print("Wavelength not in the range of the spectrum.")
            continue
        # If continuum, conv, abund or plot range to fit is smaller than 0, stop fitting
        if len(ff.cut_spec(spec_obs, lamb, cut_val=cut_val[0])) <= 1:
            print("Continuum range smaller than 0.")
            break
        elif len(ff.cut_spec(spec_obs, lamb, cut_val=cut_val[1])) <= 1:
            print("Convolution range smaller than 0.")
            break
        elif len(ff.cut_spec(spec_obs, lamb, cut_val=cut_val[2])) <= 1:
            print("Abundance range smaller than 0.")
            break
        elif len(ff.cut_spec(spec_obs, lamb, cut_val=cut_val[3])) <= 1:
            print("Plot range smaller than 0.")
            break

        if max_iter is None:
            max_iter = [100, 10]

        # arrumar segundo espectro overwriting os valores bons do primeiro
        chi, equiv_width_obs, equiv_width_fit = 0, 0, 0
        par = [abund_val_refer]
        spec_fit = [[], []]
        for repeat in range(repfit):
            # Fit of lambda shift, continuum and convolution
            spec_obs_cut = ff.cut_spec(spec_obs, lamb, cut_val=cut_val[0])

            if type_synth[0] == "Equivalent Width":
                if opt_pars is None:
                    opt_pars, chi, spec_fit = vf.optimize_spec(spec_obs_cut, type_synth, lamb, continuum,
                                                               iterac=max_iter[1], convovbound=convovbound,
                                                               wavebound=wavebound)
                if ui is not None:
                    ui.abundancelabel.setText("Deepth")
            elif type_synth[0] == "TurboSpectrum":
                tf.change_spec_range_configfl(config_fl, lamb, cut_val[0])
                tf.run_configfl(config_fl)
                spec_conv = pd.read_csv(conv_name, header=None, delimiter=r"\s+")
                if ui is not None:
                    ui.abundancelabel.setText("Abundance")

                if opt_pars is None:
                    opt_pars = tf.optimize_spec(spec_obs_cut, spec_conv, lamb, cut_val, continuum,
                                                init=opt_pars, iterac=max_iter[1], convovbound=convovbound,
                                                wavebound=wavebound)
            # opt_pars = [0,0,0]

            print("\tLamb Shift:\t\t{:.4f}\n"
                  "\tContinuum:\t\t{:.4f}\n"
                  "\tConvolution:\t\t{:.4f}".format(opt_pars[0], opt_pars[1], opt_pars[2]))
            if ui is not None:
                ui.lambshifvalue.setValue(opt_pars[0])
                ui.continuumvalue.setValue(opt_pars[1])
                ui.convolutionvalue.setValue(opt_pars[2])
                # Allow QT to actualize the UI while in the loop
                QtCore.QCoreApplication.processEvents()
                if ui.stop_state:
                    ui.stop_state = False
                    return found_val, ax, plot_line_refer

            # Fit of abundance
            spec_obs_cut = ff.cut_spec(spec_obs, lamb, cut_val=cut_val[3])
            if type_synth[0] == "Equivalent Width":
                par, chi, spec_fit = vf.optimize_abund(spec_obs_cut, type_synth, lamb, opt_pars, iterac=max_iter[2])
            elif type_synth[0] == "TurboSpectrum":
                tf.change_spec_range_configfl(config_fl, lamb, cut_val[3])

                par, chi, spec_fit = tf.optimize_abund(spec_obs_cut, config_fl, conv_name, elem, opt_pars, par,
                                                       abund_lim, iterac=max_iter[2])

            print("\tAbundance:\t\t{:.4f}".format(par[0]))

            if ui is not None:
                ui.abundancevalue.setValue(par[0])
                # Allow QT to actualize the UI while in the loop
                QtCore.QCoreApplication.processEvents()
                if ui.stop_state:
                    ui.stop_state = False
                    return found_val, ax, plot_line_refer

            # Fit of Equivalent Width Observed Spectrum
            # noinspection PyUnresolvedReferences
            spec1d = Spectrum1D(spectral_axis=np.asarray(spec_obs_cut[0]) * u.AA,
                                flux=np.asarray(spec_obs_cut[1]) * u.Jy)
            equiv_width_obs = equivalent_width(spec1d)
            # noinspection PyUnresolvedReferences
            equiv_width_obs = np.float16(equiv_width_obs / u.AA)

            # Fit of Equivalent Width Fitted Spectrum
            # noinspection PyUnresolvedReferences
            spec1d = Spectrum1D(spectral_axis=np.asarray(spec_fit[0]) * u.AA,
                                flux=np.asarray(spec_fit[1]) * u.Jy)
            equiv_width_fit = equivalent_width(spec1d)
            # noinspection PyUnresolvedReferences
            equiv_width_fit = np.float16(equiv_width_fit / u.AA)

        index_append = len(found_val) if only_abund_ind is None else only_abund_ind
        found_val.loc[index_append] = [elem+order, lamb, opt_pars[0], opt_pars[1], opt_pars[2], abund_val_refer,
                                       par[0], np.abs(par[0]-abund_val_refer), "{:.4e}".format(chi),
                                       "{:.4e}".format(equiv_width_obs), "{:.4e}".format(equiv_width_fit)]

        # Plot of data
        spec_obs_cut = ff.cut_spec(spec_obs, lamb, cut_val=cut_val[3])
        if type_synth[0] == "Equivalent Width":
            func = vf.find_func(type_synth[1])
            x = np.linspace(min(spec_obs_cut.iloc[:, 0]), max(spec_obs_cut.iloc[:, 0]), 1000)
            spec_fit = pd.DataFrame({0: x, 1: func(x, b=opt_pars[0]+lamb, c=opt_pars[2], a=par[0], d=opt_pars[1])})
        elif type_synth[0] == "TurboSpectrum":
            tf.change_abund_configfl(config_fl, elem, find=False, abund=par[0])
            tf.change_spec_range_configfl(config_fl, lamb, cut_val[3])
            tf.run_configfl(config_fl)
            # Return the Turbospectrum2019 configuration file to original abundance value
            tf.change_abund_configfl(config_fl, elem, find=False, abund=abund_val_refer)

            spec_conv = pd.read_csv(conv_name, header=None, delimiter=r"\s+")

            # noinspection PyTypeChecker
            spec_fit = ff.spec_operations(spec_conv.copy(), lamb_desloc=opt_pars[0], continuum=opt_pars[1],
                                          convol=opt_pars[2])

        if ui is None:
            plot_spec(spec_obs_cut, spec_conv, spec_fit, lamb, elem+order, folder)
        else:
            spec_fit_arr = [spec_fit]
            plot_line_refer = plot_spec_ui(spec_fit_arr, folder, elem, lamb, order, ax, canvas, plot_line_refer)

        # Write the line result to the csv file
        found_val.to_csv(Path(folder).joinpath(save_name), index=False, float_format="%.4f")

        if ui is not None:
            if only_abund_ind is None:
                rowpos = ui.abundancetable.rowCount()
                ui.abundancetable.insertRow(rowpos)
                ui.abundancetable.setItem(rowpos, 0, QtWidgets.QTableWidgetItem(str(elem)+str(order)))
                ui.abundancetable.setItem(rowpos, 1, QtWidgets.QTableWidgetItem(str(lamb)))

            # linescurrent = ui.progressvalue.text().split("/")
            ui.progressvalue.setText("{}/{}".format(i+1, len(linelist)))

    return found_val, ax, plot_line_refer


def read_config(config_name):
    """
    Function to read the configuration file (previously necessary in
    the older no-gui version).

    :param config_name: configuration file name.
    :return: linelist file name, reference abundance file name, type of the
             synthetics spectrum to use, TurboSpectrum configuration file name,
             TurboSpectrum results file name, directory to save the results,
             spectrum file name.
    """

    def separator(sep):
        # Define the type of separator
        if sep == "comma":
            return ","
        elif sep == "tab":
            return r"\s+"
        else:
            return ","

    # Read file
    config_data = pd.read_csv(config_name, comment="#", header=None)

    # Enable long strings
    pd.options.display.max_colwidth = 150

    # Attribution of the variables
    list_name = [config_data.loc[0].to_string(index=False),
                 config_data.loc[1].to_string(index=False),
                 config_data.loc[2].to_string(index=False)]
    refer_name = [config_data.loc[3].to_string(index=False),
                  config_data.loc[4].to_string(index=False)]
    type_synth = config_data.loc[5].to_string(index=False)
    config_fl = config_data.loc[6].to_string(index=False)
    conv_name = config_data.loc[7].to_string(index=False)
    folder = config_data.loc[8].to_string(index=False) + "/"

    # Attribution of the spectra names
    observed_name = []
    for i in range(9, len(config_data)):
        observed_name.append(config_data.iloc[i].to_string(index=False))

    # Define the separator type in the file
    list_name[1] = separator(list_name[1])
    refer_name[1] = separator(refer_name[1])
    for i in range(1, len(observed_name), 2):
        observed_name[i] = separator(observed_name[i])

    return list_name, refer_name, type_synth, config_fl, conv_name, folder, observed_name


def gui_call(spec_obs, ui, checkstate, canvas, ax, cut_val=None, plot_line_refer=None, opt_pars=None,
             repfit=2, abundplot=None, results_array=None, only_abund_ind=None, final_plot=False,
             plot_type=None, single=None):
    """
    Main function to be called from the GUI.

    :param spec_obs: spectrum data.
    :param ui: main GUI QT object.
    :param checkstate: QT checked state for checkboxes.
    :param canvas: GUI canvas plot object.
    :param ax: GUI ax plot object.
    :param cut_val: range to cut the spectrum.
    :param plot_line_refer: array with a list of the current available plots.
    :param opt_pars: convolution, wavelength shift and continuum parameters.
    :param repfit: number of iterations of the main fit function.
    :param abundplot: overwrite the abundance fit and only plot the value
                      present in this variable.
    :param results_array: array for the results of the fit.
    :param only_abund_ind: change only abundance without fitting other parameters at this index.
    :param final_plot: if it's meant to create the final plots instead of fitting.
    :param plot_type: type of the final plot to create.
    :param single: create a final plot to a single line or for all of them.
    :return: the results of the fit, the actualized ``ax`` variable and the
             actualized ``plot_line_refer`` variable.
    """

    # Time Counter
    init = time.time()
    init_local = time.localtime()
    init_time = "Run started at: {} {} {}, {}:{}:{}. ".format(init_local.tm_year,
                                                              init_local.tm_mon,
                                                              init_local.tm_mday,
                                                              init_local.tm_hour,
                                                              init_local.tm_min,
                                                              init_local.tm_sec)
    print(init_time)

    ui.gui_hold(True)

    if ui.restart.checkState() == QtCore.Qt.CheckState.Checked:
        restart = True
    else:
        restart = False

    folder = ui.outputname.text()

    # Create necessary folders
    folder_create = Path(folder).joinpath("On_time_Plots")
    os.makedirs(folder_create, exist_ok=True)

    linelist = []
    for line in range(ui.linelistcontent.rowCount()):
        item = ui.linelistcontent.cellWidget(line, 2).layout().itemAt(0).widget()
        if item.checkState() == checkstate and \
           ui.linelistcontent.item(line, 0) is not None and \
           ui.linelistcontent.item(line, 1) is not None:
            linelist.append([ui.linelistcontent.item(line, 0).text(), ui.linelistcontent.item(line, 1).text()])
    linelist = pd.DataFrame(data=linelist)

    refer_fl = []
    inds = []
    for line in range(ui.refercontent.rowCount()):
        if ui.refercontent.item(line, 0) is not None and \
           ui.refercontent.item(line, 1) is not None:
            inds.append(ui.refercontent.item(line, 0).text())
            refer_fl.append(ui.refercontent.item(line, 1).text())
    refer_fl = pd.DataFrame(data=refer_fl, index=inds, columns=["value"])

    methodconfig = None
    if ui.methodbox.currentText() == "Equivalent Width":
        methodconfig = [ui.methodbox.currentText(),
                        ui.ewfuncbox.currentText(),
                        ui.convinitguessvalue.value(),
                        ui.deepthinitguessvalue.value()]
    elif ui.methodbox.currentText() == "TurboSpectrum":
        methodconfig = [ui.methodbox.currentText(),
                        ui.turbospectrumoutputname.text(),
                        ui.turbospectrumconfigname.text()]
        if not os.path.isfile(methodconfig[1]) or not os.path.isfile(methodconfig[2]):
            ui.gui_hold(False)

            msg_err = "Error: TurboSpectrum files are not properly set. Run aborted.\n"

            # noinspection PyTypeChecker
            with open(str(Path(folder).joinpath("log.txt")), "a") as f:
                f.write(init_time)
                f.write(msg_err)
                f.write("\n")
                f.close()
            print(msg_err)

            ui.show_error("TurboSpectrum files are not properly set. Run aborted.")

            return results_array, ax, plot_line_refer

    if not final_plot:
        ui.methodsdatafittab.setCurrentIndex(2)

    max_iter = ui.max_iter
    convovbound = ui.convovbound
    wavebound = ui.wavebound
    contpars = ui.continuumpars

    # Allow QT to actualize the UI while in the loop
    QtCore.QCoreApplication.processEvents()

    if abundplot is None and not final_plot:
        for spec in spec_obs:
            results_array, ax, plot_line_refer = fit_abundance(linelist, spec, refer_fl, folder, methodconfig,
                                                               restart=restart, ui=ui, canvas=canvas, ax=ax,
                                                               cut_val=cut_val, plot_line_refer=plot_line_refer,
                                                               opt_pars=opt_pars, repfit=repfit, max_iter=max_iter,
                                                               convovbound=convovbound, contpars=contpars,
                                                               wavebound=wavebound, only_abund_ind=only_abund_ind)
    elif not final_plot:
        currow = ui.abundancetable.currentRow()
        currow = ui.abundancetable.rowCount() - 1 if currow == -1 else currow
        elem = ui.abundancetable.item(currow, 0).text()
        lamb = ui.abundancetable.item(currow, 1).text()
        plot_line_refer = plot_abund_nofit(elem, lamb, abundplot, refer_fl, folder, methodconfig,
                                           cut_val=cut_val, canvas=canvas, ax=ax, plot_line_refer=plot_line_refer,
                                           opt_pars=opt_pars)
    else:
        ap.folders_creation(folder)

        res_arr_copy = results_array.copy()
        elements_full = results_array.Element
        elements = []
        for i in elements_full:
            elem, order = check_order(i)
            elements.append(elem)
        res_arr_copy.Element = elements
        elements = np.unique(elements)

        if plot_type == "lines":
            ui.plotstab.setCurrentIndex(0)

            if not single:
                res_to_send = results_array
            else:
                ind = ui.abundancetable.currentRow()
                res_to_send = results_array.iloc[[ind]]

            ap.plot_lines(spec_obs,
                          res_to_send,
                          refer_fl,
                          methodconfig,
                          folder,
                          cut_val=cut_val[3],
                          abundance_shift=ui.abundshift.value(),
                          ui=ui)
        elif plot_type == "box":
            ui.plotstab.setCurrentIndex(1)
            ap.plot_abund_box(res_arr_copy, elements, folder, ui=ui)
        elif plot_type == "hist":
            ui.plotstab.setCurrentIndex(2)
            if single:
                ind = ui.abundancetable.currentRow()
                elem = ui.results_array.Element.iloc[ind]
                elem, order = check_order(elem)
                elements = [elem]
                # res_arr_copy = ui.res_arr_copy[ui.results_array.Element.str.contains(elem)]

            bins = ui.histbinsvalue.value()
            ap.plot_abund_hist(res_arr_copy, elements, folder, ui=ui, bins=bins)
            ap.plot_differ_hist(res_arr_copy, elements, folder, ui=ui, bins=bins)

    ui.gui_hold(False)

    if final_plot:
        ui.run.setDisabled(True)

    # Time Counter
    end = time.time()
    dif = end - init
    time_spent = "Time spent: {:.0f} h {:.0f} m {:.0f} s ({:.0f} s).\n".format(dif // 3600, (dif // 60) % 60,
                                                                               dif % 60, dif)

    # noinspection PyTypeChecker
    with open(str(Path(folder).joinpath("log.txt")), "a") as f:
        f.write(init_time)
        f.write(time_spent)
        f.write("\n")
        f.close()
    print(time_spent)

    return results_array, ax, plot_line_refer


def main(args):
    """
    Main subroutine to call the functions (previously necessary in the older no-gui version).

    :param args: system command line arguments.
    """

    # Time Counter
    init = time.time()

    # Arguments Menu Call
    config_name = args_menu(args)

    # Read Configuration File
    list_name, refer_name, type_synth, config_fl, conv_name, folder, observed_name = read_config(config_name)

    # Create necessary folders
    if not os.path.exists(folder):
        os.mkdir(folder)
    if not os.path.exists(folder+"On_time_Plots"):
        os.mkdir(folder+"On_time_Plots")

    # Open some files
    linelist = open_linelist_refer_fl(list_name)
    refer_fl = open_linelist_refer_fl(refer_name)
    # Function to open the observed spectrum file using PANDAS
    spec_obs = []
    for i in range(0, len(observed_name), 2):
        delimiter = observed_name[i + 1]
        spec_obs.append(pd.read_csv(observed_name[i], header=None, delimiter=delimiter))

    type_synth = ["TurboSpectrum",
                  conv_name,
                  config_fl]

    # Adjust parameters for each spectra declared
    # Caution: be aware of data overlay on the final csv file
    for spec in spec_obs:
        fit_abundance(linelist, spec, refer_fl, folder, type_synth,
                      restart=False)

    # Time Counter
    end = time.time()
    dif = end - init
    time_spent = "Time spent: {:.0f} h {:.0f} m {:.0f} s ({:.0f} s)".format(dif // 3600, (dif // 60) % 60, dif % 60,
                                                                            dif)

    # noinspection PyTypeChecker
    np.savetxt(folder+"log.txt", [time_spent], fmt="%s")
    print(time_spent)


def args_menu(args):
    """
    Run the arguments resolve and print them if necessary (previously necessary in the older no-gui version).

    :param args: system command line arguments.
    """

    if len(args) <= 1 and not any((i == "-h" or i == "--h" or i == "-help" or i == "--help") for i in args):
        if len(args) != 0:
            config_name = args[0]
        else:
            config_name = "meafs_config.txt"

        if not os.path.exists(config_name):
            sys.exit("\033[1;31mError: file {} not found.\033[m".format(config_name))

        return config_name
    else:
        help_msg = "\n\t\t\033[1;31mHelp Section\033[m\nabundance_fit.py v{}\n" \
                   "Usage: python3 abundance_fit.py [options] argument\n\n" \
                   "Written by Matheus J. Castro <https://github.com/MatheusJCastro/meafs>\nUnder MIT License.\n\n" \
                   "This program finds abundances of elements for a given spectrum.\n\n" \
                   "Argument needs to be a \".txt\" file with the MEAFS configuration.\n" \
                   "If no argument is given, the default name is \"meafs_config.txt\".\n\n" \
                   "Options are:\n -h,  -help\t|\tShow this help;\n--h, --help\t|\tShow this help;\n" \
                   "\n\nExample:\n./abundance_fit.py meafs_config.txt\n".format(__version__)
        print(help_msg)


if __name__ == '__main__':
    arg = sys.argv[1:]
    main(arg)  # call main subroutine



# Form implementation generated from reading ui file 'meafs.ui'
#
# Created by: PyQt6 UI code generator 6.7.0
#
# WARNING: Any manual changes made to this file will be lost when pyuic6 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt6 import QtCore, QtGui, QtWidgets


class Ui_MEAFS(object):
    def setupUi(self, MEAFS):
        MEAFS.setObjectName("MEAFS")
        MEAFS.resize(1203, 758)
        self.centralwidget = QtWidgets.QWidget(parent=MEAFS)
        self.centralwidget.setObjectName("centralwidget")
        self.gridLayout = QtWidgets.QGridLayout(self.centralwidget)
        self.gridLayout.setObjectName("gridLayout")
        self.methodsdatafittab = QtWidgets.QTabWidget(parent=self.centralwidget)
        self.methodsdatafittab.setObjectName("methodsdatafittab")
        self.datatab = QtWidgets.QWidget()
        self.datatab.setObjectName("datatab")
        self.gridLayout_11 = QtWidgets.QGridLayout(self.datatab)
        self.gridLayout_11.setObjectName("gridLayout_11")
        self.gridLayout_10 = QtWidgets.QGridLayout()
        self.gridLayout_10.setObjectName("gridLayout_10")
        self.dataloadtable = QtWidgets.QTableWidget(parent=self.datatab)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.dataloadtable.sizePolicy().hasHeightForWidth())
        self.dataloadtable.setSizePolicy(sizePolicy)
        self.dataloadtable.setSizeAdjustPolicy(QtWidgets.QAbstractScrollArea.SizeAdjustPolicy.AdjustIgnored)
        self.dataloadtable.setObjectName("dataloadtable")
        self.dataloadtable.setColumnCount(2)
        self.dataloadtable.setRowCount(0)
        item = QtWidgets.QTableWidgetItem()
        self.dataloadtable.setHorizontalHeaderItem(0, item)
        item = QtWidgets.QTableWidgetItem()
        self.dataloadtable.setHorizontalHeaderItem(1, item)
        self.dataloadtable.horizontalHeader().setStretchLastSection(True)
        self.dataloadtable.verticalHeader().setVisible(False)
        self.gridLayout_10.addWidget(self.dataloadtable, 0, 0, 1, 2)
        self.delimitertype = QtWidgets.QComboBox(parent=self.datatab)
        self.delimitertype.setObjectName("delimitertype")
        self.delimitertype.addItem("")
        self.delimitertype.addItem("")
        self.delimitertype.addItem("")
        self.gridLayout_10.addWidget(self.delimitertype, 1, 0, 1, 1)
        self.loaddata = QtWidgets.QPushButton(parent=self.datatab)
        self.loaddata.setObjectName("loaddata")
        self.gridLayout_10.addWidget(self.loaddata, 1, 1, 1, 1)
        self.gridLayout_11.addLayout(self.gridLayout_10, 0, 0, 1, 1)
        self.methodsdatafittab.addTab(self.datatab, "")
        self.majormethodstab = QtWidgets.QWidget()
        self.majormethodstab.setObjectName("majormethodstab")
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout(self.majormethodstab)
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.methods = QtWidgets.QVBoxLayout()
        self.methods.setObjectName("methods")
        self.methodconfiglabel = QtWidgets.QLabel(parent=self.majormethodstab)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Preferred, QtWidgets.QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.methodconfiglabel.sizePolicy().hasHeightForWidth())
        self.methodconfiglabel.setSizePolicy(sizePolicy)
        font = QtGui.QFont()
        font.setPointSize(12)
        font.setBold(True)
        font.setWeight(75)
        self.methodconfiglabel.setFont(font)
        self.methodconfiglabel.setAutoFillBackground(False)
        self.methodconfiglabel.setAlignment(QtCore.Qt.AlignmentFlag.AlignCenter)
        self.methodconfiglabel.setObjectName("methodconfiglabel")
        self.methods.addWidget(self.methodconfiglabel)
        self.methodbox = QtWidgets.QComboBox(parent=self.majormethodstab)
        self.methodbox.setObjectName("methodbox")
        self.methodbox.addItem("")
        self.methodbox.addItem("")
        self.methodbox.addItem("")
        self.methodbox.addItem("")
        self.methods.addWidget(self.methodbox)
        self.methodstab = QtWidgets.QTabWidget(parent=self.majormethodstab)
        font = QtGui.QFont()
        font.setPointSize(10)
        self.methodstab.setFont(font)
        self.methodstab.setAccessibleName("")
        self.methodstab.setObjectName("methodstab")
        self.eqwidthtab = QtWidgets.QWidget()
        self.eqwidthtab.setObjectName("eqwidthtab")
        self.gridLayout_9 = QtWidgets.QGridLayout(self.eqwidthtab)
        self.gridLayout_9.setObjectName("gridLayout_9")
        self.gridLayout_8 = QtWidgets.QGridLayout()
        self.gridLayout_8.setObjectName("gridLayout_8")
        self.convinitguesslabel = QtWidgets.QLabel(parent=self.eqwidthtab)
        self.convinitguesslabel.setObjectName("convinitguesslabel")
        self.gridLayout_8.addWidget(self.convinitguesslabel, 2, 0, 1, 1)
        self.deepthinitguesslabel = QtWidgets.QLabel(parent=self.eqwidthtab)
        self.deepthinitguesslabel.setObjectName("deepthinitguesslabel")
        self.gridLayout_8.addWidget(self.deepthinitguesslabel, 3, 0, 1, 1)
        self.deepthinitguessvalue = QtWidgets.QDoubleSpinBox(parent=self.eqwidthtab)
        self.deepthinitguessvalue.setDecimals(4)
        self.deepthinitguessvalue.setMinimum(-99.99)
        self.deepthinitguessvalue.setObjectName("deepthinitguessvalue")
        self.gridLayout_8.addWidget(self.deepthinitguessvalue, 3, 1, 1, 1)
        self.convinitguessvalue = QtWidgets.QDoubleSpinBox(parent=self.eqwidthtab)
        self.convinitguessvalue.setDecimals(4)
        self.convinitguessvalue.setMinimum(-99.99)
        self.convinitguessvalue.setObjectName("convinitguessvalue")
        self.gridLayout_8.addWidget(self.convinitguessvalue, 2, 1, 1, 1)
        self.ewfuncbox = QtWidgets.QComboBox(parent=self.eqwidthtab)
        self.ewfuncbox.setObjectName("ewfuncbox")
        self.ewfuncbox.addItem("")
        self.ewfuncbox.addItem("")
        self.ewfuncbox.addItem("")
        self.gridLayout_8.addWidget(self.ewfuncbox, 0, 1, 1, 1)
        self.initguesslabel = QtWidgets.QLabel(parent=self.eqwidthtab)
        self.initguesslabel.setAlignment(QtCore.Qt.AlignmentFlag.AlignCenter)
        self.initguesslabel.setObjectName("initguesslabel")
        self.gridLayout_8.addWidget(self.initguesslabel, 1, 0, 1, 2)
        self.ewfunclabel = QtWidgets.QLabel(parent=self.eqwidthtab)
        self.ewfunclabel.setObjectName("ewfunclabel")
        self.gridLayout_8.addWidget(self.ewfunclabel, 0, 0, 1, 1)
        self.gridLayout_9.addLayout(self.gridLayout_8, 0, 0, 1, 1)
        self.methodstab.addTab(self.eqwidthtab, "")
        self.turbspectrumtab = QtWidgets.QWidget()
        self.turbspectrumtab.setObjectName("turbspectrumtab")
        self.gridLayout_3 = QtWidgets.QGridLayout(self.turbspectrumtab)
        self.gridLayout_3.setObjectName("gridLayout_3")
        self.gridLayout_2 = QtWidgets.QGridLayout()
        self.gridLayout_2.setObjectName("gridLayout_2")
        self.turbospectrumconfigname = QtWidgets.QLineEdit(parent=self.turbspectrumtab)
        self.turbospectrumconfigname.setObjectName("turbospectrumconfigname")
        self.gridLayout_2.addWidget(self.turbospectrumconfigname, 1, 0, 1, 1)
        self.turbospectrumoutputname = QtWidgets.QLineEdit(parent=self.turbspectrumtab)
        self.turbospectrumoutputname.setObjectName("turbospectrumoutputname")
        self.gridLayout_2.addWidget(self.turbospectrumoutputname, 3, 0, 1, 1)
        self.turbospectrumconfigbrowse = QtWidgets.QPushButton(parent=self.turbspectrumtab)
        self.turbospectrumconfigbrowse.setObjectName("turbospectrumconfigbrowse")
        self.gridLayout_2.addWidget(self.turbospectrumconfigbrowse, 1, 1, 1, 1)
        self.turbospectrumoutputbrowse = QtWidgets.QPushButton(parent=self.turbspectrumtab)
        self.turbospectrumoutputbrowse.setObjectName("turbospectrumoutputbrowse")
        self.gridLayout_2.addWidget(self.turbospectrumoutputbrowse, 3, 1, 1, 1)
        self.turbospectrumoutputlabel = QtWidgets.QLabel(parent=self.turbspectrumtab)
        font = QtGui.QFont()
        font.setPointSize(10)
        font.setBold(False)
        font.setWeight(50)
        self.turbospectrumoutputlabel.setFont(font)
        self.turbospectrumoutputlabel.setObjectName("turbospectrumoutputlabel")
        self.gridLayout_2.addWidget(self.turbospectrumoutputlabel, 2, 0, 1, 2)
        self.turbospectrumconfiglabel = QtWidgets.QLabel(parent=self.turbspectrumtab)
        font = QtGui.QFont()
        font.setPointSize(10)
        font.setBold(False)
        font.setWeight(50)
        self.turbospectrumconfiglabel.setFont(font)
        self.turbospectrumconfiglabel.setObjectName("turbospectrumconfiglabel")
        self.gridLayout_2.addWidget(self.turbospectrumconfiglabel, 0, 0, 1, 2)
        self.gridLayout_3.addLayout(self.gridLayout_2, 0, 0, 1, 1)
        self.methodstab.addTab(self.turbspectrumtab, "")
        self.pfanttab = QtWidgets.QWidget()
        self.pfanttab.setObjectName("pfanttab")
        self.gridLayout_7 = QtWidgets.QGridLayout(self.pfanttab)
        self.gridLayout_7.setObjectName("gridLayout_7")
        self.gridLayout_6 = QtWidgets.QGridLayout()
        self.gridLayout_6.setObjectName("gridLayout_6")
        self.label_3 = QtWidgets.QLabel(parent=self.pfanttab)
        font = QtGui.QFont()
        font.setPointSize(12)
        font.setBold(True)
        font.setWeight(75)
        self.label_3.setFont(font)
        self.label_3.setObjectName("label_3")
        self.gridLayout_6.addWidget(self.label_3, 0, 0, 1, 1)
        self.gridLayout_7.addLayout(self.gridLayout_6, 0, 0, 1, 1)
        self.methodstab.addTab(self.pfanttab, "")
        self.synthetab = QtWidgets.QWidget()
        self.synthetab.setObjectName("synthetab")
        self.gridLayout_5 = QtWidgets.QGridLayout(self.synthetab)
        self.gridLayout_5.setObjectName("gridLayout_5")
        self.gridLayout_4 = QtWidgets.QGridLayout()
        self.gridLayout_4.setObjectName("gridLayout_4")
        self.label_4 = QtWidgets.QLabel(parent=self.synthetab)
        font = QtGui.QFont()
        font.setPointSize(12)
        font.setBold(True)
        font.setWeight(75)
        self.label_4.setFont(font)
        self.label_4.setObjectName("label_4")
        self.gridLayout_4.addWidget(self.label_4, 0, 0, 1, 1)
        self.gridLayout_5.addLayout(self.gridLayout_4, 0, 0, 1, 1)
        self.methodstab.addTab(self.synthetab, "")
        self.methods.addWidget(self.methodstab)
        self.horizontalLayout_2.addLayout(self.methods)
        self.methodsdatafittab.addTab(self.majormethodstab, "")
        self.fitparameterstab = QtWidgets.QWidget()
        self.fitparameterstab.setObjectName("fitparameterstab")
        self.gridLayout_13 = QtWidgets.QGridLayout(self.fitparameterstab)
        self.gridLayout_13.setObjectName("gridLayout_13")
        self.fitparameters = QtWidgets.QGridLayout()
        self.fitparameters.setObjectName("fitparameters")
        self.line_4 = QtWidgets.QFrame(parent=self.fitparameterstab)
        self.line_4.setFrameShape(QtWidgets.QFrame.Shape.VLine)
        self.line_4.setFrameShadow(QtWidgets.QFrame.Shadow.Sunken)
        self.line_4.setObjectName("line_4")
        self.fitparameters.addWidget(self.line_4, 0, 1, 1, 1)
        self.manualfitbutton = QtWidgets.QPushButton(parent=self.fitparameterstab)
        self.manualfitbutton.setObjectName("manualfitbutton")
        self.fitparameters.addWidget(self.manualfitbutton, 5, 0, 1, 3)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setSizeConstraint(QtWidgets.QLayout.SizeConstraint.SetDefaultConstraint)
        self.horizontalLayout.setSpacing(6)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.currentvaluesplotbutton = QtWidgets.QPushButton(parent=self.fitparameterstab)
        self.currentvaluesplotbutton.setObjectName("currentvaluesplotbutton")
        self.horizontalLayout.addWidget(self.currentvaluesplotbutton)
        self.currentvaluessavebutton = QtWidgets.QPushButton(parent=self.fitparameterstab)
        self.currentvaluessavebutton.setObjectName("currentvaluessavebutton")
        self.horizontalLayout.addWidget(self.currentvaluessavebutton)
        self.fitparameters.addLayout(self.horizontalLayout, 7, 0, 1, 3)
        self.abundancevalue = QtWidgets.QDoubleSpinBox(parent=self.fitparameterstab)
        self.abundancevalue.setDecimals(4)
        self.abundancevalue.setMinimum(-1000.0)
        self.abundancevalue.setMaximum(1000.0)
        self.abundancevalue.setSingleStep(0.01)
        self.abundancevalue.setObjectName("abundancevalue")
        self.fitparameters.addWidget(self.abundancevalue, 6, 2, 1, 1)
        self.linedeflabel = QtWidgets.QLabel(parent=self.fitparameterstab)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Fixed, QtWidgets.QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.linedeflabel.sizePolicy().hasHeightForWidth())
        self.linedeflabel.setSizePolicy(sizePolicy)
        self.linedeflabel.setObjectName("linedeflabel")
        self.fitparameters.addWidget(self.linedeflabel, 0, 0, 1, 1)
        self.continuumvalue = QtWidgets.QDoubleSpinBox(parent=self.fitparameterstab)
        self.continuumvalue.setDecimals(4)
        self.continuumvalue.setMinimum(-1000.0)
        self.continuumvalue.setMaximum(1000.0)
        self.continuumvalue.setSingleStep(0.001)
        self.continuumvalue.setObjectName("continuumvalue")
        self.fitparameters.addWidget(self.continuumvalue, 3, 2, 1, 1)
        self.convolutionlabel = QtWidgets.QLabel(parent=self.fitparameterstab)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Fixed, QtWidgets.QSizePolicy.Policy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.convolutionlabel.sizePolicy().hasHeightForWidth())
        self.convolutionlabel.setSizePolicy(sizePolicy)
        self.convolutionlabel.setObjectName("convolutionlabel")
        self.fitparameters.addWidget(self.convolutionlabel, 4, 0, 1, 1)
        self.continuumlabel = QtWidgets.QLabel(parent=self.fitparameterstab)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Fixed, QtWidgets.QSizePolicy.Policy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.continuumlabel.sizePolicy().hasHeightForWidth())
        self.continuumlabel.setSizePolicy(sizePolicy)
        self.continuumlabel.setObjectName("continuumlabel")
        self.fitparameters.addWidget(self.continuumlabel, 3, 0, 1, 1)
        self.linedefvalue = QtWidgets.QLabel(parent=self.fitparameterstab)
        self.linedefvalue.setText("")
        self.linedefvalue.setObjectName("linedefvalue")
        self.fitparameters.addWidget(self.linedefvalue, 0, 2, 1, 1)
        self.abundancelabel = QtWidgets.QLabel(parent=self.fitparameterstab)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Fixed, QtWidgets.QSizePolicy.Policy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.abundancelabel.sizePolicy().hasHeightForWidth())
        self.abundancelabel.setSizePolicy(sizePolicy)
        self.abundancelabel.setObjectName("abundancelabel")
        self.fitparameters.addWidget(self.abundancelabel, 6, 0, 1, 1)
        self.lambshiflabel = QtWidgets.QLabel(parent=self.fitparameterstab)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Fixed, QtWidgets.QSizePolicy.Policy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.lambshiflabel.sizePolicy().hasHeightForWidth())
        self.lambshiflabel.setSizePolicy(sizePolicy)
        self.lambshiflabel.setObjectName("lambshiflabel")
        self.fitparameters.addWidget(self.lambshiflabel, 2, 0, 1, 1)
        self.lambshifvalue = QtWidgets.QDoubleSpinBox(parent=self.fitparameterstab)
        self.lambshifvalue.setDecimals(4)
        self.lambshifvalue.setMinimum(-1000.0)
        self.lambshifvalue.setMaximum(1000.0)
        self.lambshifvalue.setSingleStep(0.001)
        self.lambshifvalue.setObjectName("lambshifvalue")
        self.fitparameters.addWidget(self.lambshifvalue, 2, 2, 1, 1)
        self.line = QtWidgets.QFrame(parent=self.fitparameterstab)
        self.line.setFrameShape(QtWidgets.QFrame.Shape.VLine)
        self.line.setFrameShadow(QtWidgets.QFrame.Shadow.Sunken)
        self.line.setObjectName("line")
        self.fitparameters.addWidget(self.line, 2, 1, 3, 1)
        self.convolutionvalue = QtWidgets.QDoubleSpinBox(parent=self.fitparameterstab)
        self.convolutionvalue.setDecimals(4)
        self.convolutionvalue.setMinimum(-1000.0)
        self.convolutionvalue.setMaximum(1000.0)
        self.convolutionvalue.setSingleStep(0.001)
        self.convolutionvalue.setObjectName("convolutionvalue")
        self.fitparameters.addWidget(self.convolutionvalue, 4, 2, 1, 1)
        self.line_3 = QtWidgets.QFrame(parent=self.fitparameterstab)
        self.line_3.setFrameShape(QtWidgets.QFrame.Shape.VLine)
        self.line_3.setFrameShadow(QtWidgets.QFrame.Shadow.Sunken)
        self.line_3.setObjectName("line_3")
        self.fitparameters.addWidget(self.line_3, 6, 1, 1, 1)
        self.line_2 = QtWidgets.QFrame(parent=self.fitparameterstab)
        self.line_2.setFrameShape(QtWidgets.QFrame.Shape.HLine)
        self.line_2.setFrameShadow(QtWidgets.QFrame.Shadow.Sunken)
        self.line_2.setObjectName("line_2")
        self.fitparameters.addWidget(self.line_2, 1, 0, 1, 3)
        self.gridLayout_13.addLayout(self.fitparameters, 0, 0, 1, 1)
        self.methodsdatafittab.addTab(self.fitparameterstab, "")
        self.finalplotsruntab = QtWidgets.QWidget()
        self.finalplotsruntab.setObjectName("finalplotsruntab")
        self.gridLayout_17 = QtWidgets.QGridLayout(self.finalplotsruntab)
        self.gridLayout_17.setObjectName("gridLayout_17")
        self.gridLayout_15 = QtWidgets.QGridLayout()
        self.gridLayout_15.setObjectName("gridLayout_15")
        self.linesplotlabel = QtWidgets.QLabel(parent=self.finalplotsruntab)
        self.linesplotlabel.setAlignment(QtCore.Qt.AlignmentFlag.AlignCenter)
        self.linesplotlabel.setObjectName("linesplotlabel")
        self.gridLayout_15.addWidget(self.linesplotlabel, 0, 0, 1, 2)
        self.histbinslabel = QtWidgets.QLabel(parent=self.finalplotsruntab)
        self.histbinslabel.setObjectName("histbinslabel")
        self.gridLayout_15.addWidget(self.histbinslabel, 10, 0, 1, 1)
        self.line_5 = QtWidgets.QFrame(parent=self.finalplotsruntab)
        self.line_5.setFrameShape(QtWidgets.QFrame.Shape.HLine)
        self.line_5.setFrameShadow(QtWidgets.QFrame.Shadow.Sunken)
        self.line_5.setObjectName("line_5")
        self.gridLayout_15.addWidget(self.line_5, 5, 0, 1, 2)
        self.line_6 = QtWidgets.QFrame(parent=self.finalplotsruntab)
        self.line_6.setFrameShape(QtWidgets.QFrame.Shape.HLine)
        self.line_6.setFrameShadow(QtWidgets.QFrame.Shadow.Sunken)
        self.line_6.setObjectName("line_6")
        self.gridLayout_15.addWidget(self.line_6, 8, 0, 1, 2)
        self.boxplotcreate = QtWidgets.QPushButton(parent=self.finalplotsruntab)
        self.boxplotcreate.setObjectName("boxplotcreate")
        self.gridLayout_15.addWidget(self.boxplotcreate, 7, 0, 1, 2)
        self.boxplotlabel = QtWidgets.QLabel(parent=self.finalplotsruntab)
        self.boxplotlabel.setAlignment(QtCore.Qt.AlignmentFlag.AlignCenter)
        self.boxplotlabel.setObjectName("boxplotlabel")
        self.gridLayout_15.addWidget(self.boxplotlabel, 6, 0, 1, 2)
        self.histbinsvalue = QtWidgets.QSpinBox(parent=self.finalplotsruntab)
        self.histbinsvalue.setMaximum(1000)
        self.histbinsvalue.setProperty("value", 20)
        self.histbinsvalue.setObjectName("histbinsvalue")
        self.gridLayout_15.addWidget(self.histbinsvalue, 10, 1, 1, 1)
        self.histplotlabel = QtWidgets.QLabel(parent=self.finalplotsruntab)
        self.histplotlabel.setAlignment(QtCore.Qt.AlignmentFlag.AlignCenter)
        self.histplotlabel.setObjectName("histplotlabel")
        self.gridLayout_15.addWidget(self.histplotlabel, 9, 0, 1, 2)
        self.linesplotall = QtWidgets.QPushButton(parent=self.finalplotsruntab)
        self.linesplotall.setObjectName("linesplotall")
        self.gridLayout_15.addWidget(self.linesplotall, 3, 0, 1, 1)
        self.abundshift = QtWidgets.QDoubleSpinBox(parent=self.finalplotsruntab)
        self.abundshift.setMaximum(100.0)
        self.abundshift.setSingleStep(0.1)
        self.abundshift.setProperty("value", 0.1)
        self.abundshift.setObjectName("abundshift")
        self.gridLayout_15.addWidget(self.abundshift, 2, 1, 1, 1)
        self.abundshiftlabel = QtWidgets.QLabel(parent=self.finalplotsruntab)
        self.abundshiftlabel.setObjectName("abundshiftlabel")
        self.gridLayout_15.addWidget(self.abundshiftlabel, 2, 0, 1, 1)
        self.histplotsingle = QtWidgets.QPushButton(parent=self.finalplotsruntab)
        self.histplotsingle.setObjectName("histplotsingle")
        self.gridLayout_15.addWidget(self.histplotsingle, 11, 1, 1, 1)
        self.linesplotsingle = QtWidgets.QPushButton(parent=self.finalplotsruntab)
        self.linesplotsingle.setObjectName("linesplotsingle")
        self.gridLayout_15.addWidget(self.linesplotsingle, 3, 1, 1, 1)
        self.hitsplotall = QtWidgets.QPushButton(parent=self.finalplotsruntab)
        self.hitsplotall.setObjectName("hitsplotall")
        self.gridLayout_15.addWidget(self.hitsplotall, 11, 0, 1, 1)
        self.gridLayout_17.addLayout(self.gridLayout_15, 0, 0, 1, 1)
        self.methodsdatafittab.addTab(self.finalplotsruntab, "")
        self.gridLayout.addWidget(self.methodsdatafittab, 4, 2, 1, 1)
        self.tabplotshels = QtWidgets.QTabWidget(parent=self.centralwidget)
        self.tabplotshels.setObjectName("tabplotshels")
        self.tabplot = QtWidgets.QWidget()
        self.tabplot.setObjectName("tabplot")
        self.horizontalLayout_4 = QtWidgets.QHBoxLayout(self.tabplot)
        self.horizontalLayout_4.setObjectName("horizontalLayout_4")
        self.plot = QtWidgets.QHBoxLayout()
        self.plot.setObjectName("plot")
        self.horizontalLayout_4.addLayout(self.plot)
        self.tabplotshels.addTab(self.tabplot, "")
        self.tabfinalplots = QtWidgets.QWidget()
        self.tabfinalplots.setObjectName("tabfinalplots")
        self.gridLayout_14 = QtWidgets.QGridLayout(self.tabfinalplots)
        self.gridLayout_14.setObjectName("gridLayout_14")
        self.plotstab = QtWidgets.QTabWidget(parent=self.tabfinalplots)
        font = QtGui.QFont()
        font.setPointSize(10)
        self.plotstab.setFont(font)
        self.plotstab.setLayoutDirection(QtCore.Qt.LayoutDirection.LeftToRight)
        self.plotstab.setAutoFillBackground(False)
        self.plotstab.setTabPosition(QtWidgets.QTabWidget.TabPosition.East)
        self.plotstab.setTabShape(QtWidgets.QTabWidget.TabShape.Rounded)
        self.plotstab.setElideMode(QtCore.Qt.TextElideMode.ElideNone)
        self.plotstab.setDocumentMode(True)
        self.plotstab.setTabsClosable(False)
        self.plotstab.setMovable(False)
        self.plotstab.setTabBarAutoHide(False)
        self.plotstab.setObjectName("plotstab")
        self.linesplottab = QtWidgets.QWidget()
        self.linesplottab.setObjectName("linesplottab")
        self.horizontalLayout_3 = QtWidgets.QHBoxLayout(self.linesplottab)
        self.horizontalLayout_3.setObjectName("horizontalLayout_3")
        self.linesplotimage = QtWidgets.QLabel(parent=self.linesplottab)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Preferred, QtWidgets.QSizePolicy.Policy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.linesplotimage.sizePolicy().hasHeightForWidth())
        self.linesplotimage.setSizePolicy(sizePolicy)
        font = QtGui.QFont()
        font.setPointSize(15)
        font.setBold(True)
        font.setWeight(75)
        self.linesplotimage.setFont(font)
        self.linesplotimage.setAlignment(QtCore.Qt.AlignmentFlag.AlignCenter)
        self.linesplotimage.setObjectName("linesplotimage")
        self.horizontalLayout_3.addWidget(self.linesplotimage)
        self.plotstab.addTab(self.linesplottab, "")
        self.boxplottab = QtWidgets.QWidget()
        self.boxplottab.setObjectName("boxplottab")
        self.horizontalLayout_6 = QtWidgets.QHBoxLayout(self.boxplottab)
        self.horizontalLayout_6.setObjectName("horizontalLayout_6")
        self.boxplot = QtWidgets.QHBoxLayout()
        self.boxplot.setObjectName("boxplot")
        self.boxplotimage = QtWidgets.QLabel(parent=self.boxplottab)
        font = QtGui.QFont()
        font.setPointSize(15)
        font.setBold(True)
        font.setWeight(75)
        self.boxplotimage.setFont(font)
        self.boxplotimage.setAlignment(QtCore.Qt.AlignmentFlag.AlignCenter)
        self.boxplotimage.setObjectName("boxplotimage")
        self.boxplot.addWidget(self.boxplotimage)
        self.horizontalLayout_6.addLayout(self.boxplot)
        self.plotstab.addTab(self.boxplottab, "")
        self.histplottabs = QtWidgets.QWidget()
        self.histplottabs.setObjectName("histplottabs")
        self.gridLayout_18 = QtWidgets.QGridLayout(self.histplottabs)
        self.gridLayout_18.setObjectName("gridLayout_18")
        self.histplottab = QtWidgets.QTabWidget(parent=self.histplottabs)
        self.histplottab.setLayoutDirection(QtCore.Qt.LayoutDirection.LeftToRight)
        self.histplottab.setTabPosition(QtWidgets.QTabWidget.TabPosition.East)
        self.histplottab.setTabShape(QtWidgets.QTabWidget.TabShape.Rounded)
        self.histplottab.setDocumentMode(True)
        self.histplottab.setObjectName("histplottab")
        self.abundhisttab = QtWidgets.QWidget()
        self.abundhisttab.setObjectName("abundhisttab")
        self.gridLayout_19 = QtWidgets.QGridLayout(self.abundhisttab)
        self.gridLayout_19.setObjectName("gridLayout_19")
        self.abundhistplotimage = QtWidgets.QLabel(parent=self.abundhisttab)
        font = QtGui.QFont()
        font.setPointSize(15)
        font.setBold(True)
        font.setWeight(75)
        self.abundhistplotimage.setFont(font)
        self.abundhistplotimage.setAlignment(QtCore.Qt.AlignmentFlag.AlignCenter)
        self.abundhistplotimage.setObjectName("abundhistplotimage")
        self.gridLayout_19.addWidget(self.abundhistplotimage, 0, 0, 1, 1)
        self.histplottab.addTab(self.abundhisttab, "")
        self.differhisttab = QtWidgets.QWidget()
        self.differhisttab.setObjectName("differhisttab")
        self.gridLayout_20 = QtWidgets.QGridLayout(self.differhisttab)
        self.gridLayout_20.setObjectName("gridLayout_20")
        self.differhistplotimage = QtWidgets.QLabel(parent=self.differhisttab)
        font = QtGui.QFont()
        font.setPointSize(15)
        font.setBold(True)
        font.setWeight(75)
        self.differhistplotimage.setFont(font)
        self.differhistplotimage.setAlignment(QtCore.Qt.AlignmentFlag.AlignCenter)
        self.differhistplotimage.setObjectName("differhistplotimage")
        self.gridLayout_20.addWidget(self.differhistplotimage, 0, 0, 1, 1)
        self.histplottab.addTab(self.differhisttab, "")
        self.gridLayout_18.addWidget(self.histplottab, 0, 0, 1, 1)
        self.plotstab.addTab(self.histplottabs, "")
        self.gridLayout_14.addWidget(self.plotstab, 0, 0, 1, 1)
        self.tabplotshels.addTab(self.tabfinalplots, "")
        self.tabjupyshell = QtWidgets.QWidget()
        self.tabjupyshell.setObjectName("tabjupyshell")
        self.gridLayout_16 = QtWidgets.QGridLayout(self.tabjupyshell)
        self.gridLayout_16.setObjectName("gridLayout_16")
        self.layoutjupyshell = QtWidgets.QGridLayout()
        self.layoutjupyshell.setObjectName("layoutjupyshell")
        self.gridLayout_16.addLayout(self.layoutjupyshell, 0, 0, 1, 1)
        self.tabplotshels.addTab(self.tabjupyshell, "")
        self.tabstoutstderr = QtWidgets.QWidget()
        self.tabstoutstderr.setObjectName("tabstoutstderr")
        self.gridLayout_12 = QtWidgets.QGridLayout(self.tabstoutstderr)
        self.gridLayout_12.setObjectName("gridLayout_12")
        self.clearstdbutton = QtWidgets.QToolButton(parent=self.tabstoutstderr)
        self.clearstdbutton.setObjectName("clearstdbutton")
        self.gridLayout_12.addWidget(self.clearstdbutton, 2, 1, 1, 1)
        self.stdtext = QtWidgets.QTextEdit(parent=self.tabstoutstderr)
        self.stdtext.setObjectName("stdtext")
        self.gridLayout_12.addWidget(self.stdtext, 1, 0, 1, 2)
        self.tabplotshels.addTab(self.tabstoutstderr, "")
        self.gridLayout.addWidget(self.tabplotshels, 3, 1, 1, 2)
        self.linelist = QtWidgets.QGroupBox(parent=self.centralwidget)
        self.linelist.setObjectName("linelist")
        self.linelist2 = QtWidgets.QGridLayout(self.linelist)
        self.linelist2.setObjectName("linelist2")
        self.linelistlabel = QtWidgets.QLabel(parent=self.linelist)
        font = QtGui.QFont()
        font.setPointSize(12)
        font.setBold(True)
        font.setWeight(75)
        self.linelistlabel.setFont(font)
        self.linelistlabel.setFrameShape(QtWidgets.QFrame.Shape.NoFrame)
        self.linelistlabel.setTextFormat(QtCore.Qt.TextFormat.PlainText)
        self.linelistlabel.setScaledContents(False)
        self.linelistlabel.setAlignment(QtCore.Qt.AlignmentFlag.AlignCenter)
        self.linelistlabel.setTextInteractionFlags(QtCore.Qt.TextInteractionFlag.NoTextInteraction)
        self.linelistlabel.setObjectName("linelistlabel")
        self.linelist2.addWidget(self.linelistlabel, 0, 0, 1, 5)
        self.linelistload = QtWidgets.QPushButton(parent=self.linelist)
        self.linelistload.setObjectName("linelistload")
        self.linelist2.addWidget(self.linelistload, 1, 4, 1, 1)
        self.linelistname = QtWidgets.QLineEdit(parent=self.linelist)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.linelistname.sizePolicy().hasHeightForWidth())
        self.linelistname.setSizePolicy(sizePolicy)
        self.linelistname.setObjectName("linelistname")
        self.linelist2.addWidget(self.linelistname, 1, 1, 1, 1)
        self.linelistbrowse = QtWidgets.QPushButton(parent=self.linelist)
        self.linelistbrowse.setObjectName("linelistbrowse")
        self.linelist2.addWidget(self.linelistbrowse, 1, 2, 1, 1)
        self.linelistlabel2 = QtWidgets.QLabel(parent=self.linelist)
        self.linelistlabel2.setEnabled(True)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Fixed, QtWidgets.QSizePolicy.Policy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.linelistlabel2.sizePolicy().hasHeightForWidth())
        self.linelistlabel2.setSizePolicy(sizePolicy)
        font = QtGui.QFont()
        font.setBold(False)
        font.setWeight(50)
        self.linelistlabel2.setFont(font)
        self.linelistlabel2.setObjectName("linelistlabel2")
        self.linelist2.addWidget(self.linelistlabel2, 1, 0, 1, 1)
        self.verticalLayout = QtWidgets.QVBoxLayout()
        self.verticalLayout.setObjectName("verticalLayout")
        self.linelistcheckbox = QtWidgets.QCheckBox(parent=self.linelist)
        self.linelistcheckbox.setEnabled(False)
        self.linelistcheckbox.setLayoutDirection(QtCore.Qt.LayoutDirection.LeftToRight)
        self.linelistcheckbox.setObjectName("linelistcheckbox")
        self.verticalLayout.addWidget(self.linelistcheckbox)
        self.linelistaddline = QtWidgets.QPushButton(parent=self.linelist)
        self.linelistaddline.setObjectName("linelistaddline")
        self.verticalLayout.addWidget(self.linelistaddline)
        self.linelistremoveline = QtWidgets.QPushButton(parent=self.linelist)
        self.linelistremoveline.setObjectName("linelistremoveline")
        self.verticalLayout.addWidget(self.linelistremoveline)
        self.restart = QtWidgets.QCheckBox(parent=self.linelist)
        self.restart.setLayoutDirection(QtCore.Qt.LayoutDirection.LeftToRight)
        self.restart.setObjectName("restart")
        self.verticalLayout.addWidget(self.restart)
        self.linelist2.addLayout(self.verticalLayout, 3, 4, 1, 1)
        self.linelistcontent = QtWidgets.QTableWidget(parent=self.linelist)
        self.linelistcontent.setSizeAdjustPolicy(QtWidgets.QAbstractScrollArea.SizeAdjustPolicy.AdjustIgnored)
        self.linelistcontent.setObjectName("linelistcontent")
        self.linelistcontent.setColumnCount(3)
        self.linelistcontent.setRowCount(0)
        item = QtWidgets.QTableWidgetItem()
        self.linelistcontent.setHorizontalHeaderItem(0, item)
        item = QtWidgets.QTableWidgetItem()
        self.linelistcontent.setHorizontalHeaderItem(1, item)
        item = QtWidgets.QTableWidgetItem()
        self.linelistcontent.setHorizontalHeaderItem(2, item)
        self.linelistcontent.horizontalHeader().setStretchLastSection(True)
        self.linelist2.addWidget(self.linelistcontent, 3, 0, 1, 4)
        self.gridLayout.addWidget(self.linelist, 3, 3, 1, 1)
        self.refer = QtWidgets.QGroupBox(parent=self.centralwidget)
        self.refer.setObjectName("refer")
        self.refer2 = QtWidgets.QGridLayout(self.refer)
        self.refer2.setObjectName("refer2")
        self.refername = QtWidgets.QLineEdit(parent=self.refer)
        self.refername.setObjectName("refername")
        self.refer2.addWidget(self.refername, 1, 1, 1, 1)
        self.referlabel = QtWidgets.QLabel(parent=self.refer)
        font = QtGui.QFont()
        font.setPointSize(12)
        font.setBold(True)
        font.setWeight(75)
        self.referlabel.setFont(font)
        self.referlabel.setAlignment(QtCore.Qt.AlignmentFlag.AlignCenter)
        self.referlabel.setObjectName("referlabel")
        self.refer2.addWidget(self.referlabel, 0, 0, 1, 4)
        self.referlabel2 = QtWidgets.QLabel(parent=self.refer)
        font = QtGui.QFont()
        font.setBold(False)
        font.setWeight(50)
        self.referlabel2.setFont(font)
        self.referlabel2.setObjectName("referlabel2")
        self.refer2.addWidget(self.referlabel2, 1, 0, 1, 1)
        self.verticalLayout_3 = QtWidgets.QVBoxLayout()
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.referaddline = QtWidgets.QPushButton(parent=self.refer)
        self.referaddline.setObjectName("referaddline")
        self.verticalLayout_3.addWidget(self.referaddline)
        self.referremoveline = QtWidgets.QPushButton(parent=self.refer)
        self.referremoveline.setObjectName("referremoveline")
        self.verticalLayout_3.addWidget(self.referremoveline)
        self.refer2.addLayout(self.verticalLayout_3, 2, 3, 1, 1)
        self.referbrowse = QtWidgets.QPushButton(parent=self.refer)
        self.referbrowse.setObjectName("referbrowse")
        self.refer2.addWidget(self.referbrowse, 1, 2, 1, 1)
        self.referload = QtWidgets.QPushButton(parent=self.refer)
        self.referload.setObjectName("referload")
        self.refer2.addWidget(self.referload, 1, 3, 1, 1)
        self.refercontent = QtWidgets.QTableWidget(parent=self.refer)
        self.refercontent.setObjectName("refercontent")
        self.refercontent.setColumnCount(2)
        self.refercontent.setRowCount(0)
        item = QtWidgets.QTableWidgetItem()
        self.refercontent.setHorizontalHeaderItem(0, item)
        item = QtWidgets.QTableWidgetItem()
        self.refercontent.setHorizontalHeaderItem(1, item)
        self.refercontent.horizontalHeader().setVisible(True)
        self.refercontent.horizontalHeader().setCascadingSectionResizes(False)
        self.refercontent.horizontalHeader().setSortIndicatorShown(False)
        self.refercontent.horizontalHeader().setStretchLastSection(True)
        self.refer2.addWidget(self.refercontent, 2, 0, 1, 3)
        self.gridLayout.addWidget(self.refer, 4, 3, 1, 1)
        self.outputabundances = QtWidgets.QGroupBox(parent=self.centralwidget)
        self.outputabundances.setObjectName("outputabundances")
        self.outputabundances2 = QtWidgets.QGridLayout(self.outputabundances)
        self.outputabundances2.setObjectName("outputabundances2")
        self.outputbrowse = QtWidgets.QPushButton(parent=self.outputabundances)
        self.outputbrowse.setObjectName("outputbrowse")
        self.outputabundances2.addWidget(self.outputbrowse, 3, 4, 1, 1)
        self.outputname = QtWidgets.QLineEdit(parent=self.outputabundances)
        self.outputname.setObjectName("outputname")
        self.outputabundances2.addWidget(self.outputname, 3, 3, 1, 1)
        self.abundanceslabel = QtWidgets.QLabel(parent=self.outputabundances)
        font = QtGui.QFont()
        font.setPointSize(12)
        font.setBold(True)
        font.setWeight(75)
        self.abundanceslabel.setFont(font)
        self.abundanceslabel.setAlignment(QtCore.Qt.AlignmentFlag.AlignCenter)
        self.abundanceslabel.setObjectName("abundanceslabel")
        self.outputabundances2.addWidget(self.abundanceslabel, 5, 1, 1, 4)
        self.abundancetable = QtWidgets.QTableWidget(parent=self.outputabundances)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.abundancetable.sizePolicy().hasHeightForWidth())
        self.abundancetable.setSizePolicy(sizePolicy)
        self.abundancetable.setObjectName("abundancetable")
        self.abundancetable.setColumnCount(2)
        self.abundancetable.setRowCount(0)
        item = QtWidgets.QTableWidgetItem()
        self.abundancetable.setHorizontalHeaderItem(0, item)
        item = QtWidgets.QTableWidgetItem()
        self.abundancetable.setHorizontalHeaderItem(1, item)
        self.abundancetable.horizontalHeader().setStretchLastSection(True)
        self.outputabundances2.addWidget(self.abundancetable, 6, 0, 1, 5)
        self.stop = QtWidgets.QPushButton(parent=self.outputabundances)
        font = QtGui.QFont()
        font.setPointSize(12)
        font.setBold(True)
        font.setWeight(75)
        self.stop.setFont(font)
        self.stop.setObjectName("stop")
        self.outputabundances2.addWidget(self.stop, 4, 4, 1, 1)
        self.progresslabel = QtWidgets.QLabel(parent=self.outputabundances)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Fixed, QtWidgets.QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.progresslabel.sizePolicy().hasHeightForWidth())
        self.progresslabel.setSizePolicy(sizePolicy)
        self.progresslabel.setObjectName("progresslabel")
        self.outputabundances2.addWidget(self.progresslabel, 4, 1, 1, 1)
        self.run = QtWidgets.QPushButton(parent=self.outputabundances)
        font = QtGui.QFont()
        font.setPointSize(12)
        font.setBold(True)
        font.setWeight(75)
        self.run.setFont(font)
        self.run.setObjectName("run")
        self.outputabundances2.addWidget(self.run, 4, 3, 1, 1)
        self.progressvalue = QtWidgets.QLabel(parent=self.outputabundances)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Preferred, QtWidgets.QSizePolicy.Policy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.progressvalue.sizePolicy().hasHeightForWidth())
        self.progressvalue.setSizePolicy(sizePolicy)
        self.progressvalue.setObjectName("progressvalue")
        self.outputabundances2.addWidget(self.progressvalue, 4, 2, 1, 1)
        self.outputlabel = QtWidgets.QLabel(parent=self.outputabundances)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Fixed, QtWidgets.QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.outputlabel.sizePolicy().hasHeightForWidth())
        self.outputlabel.setSizePolicy(sizePolicy)
        font = QtGui.QFont()
        font.setBold(False)
        font.setWeight(50)
        self.outputlabel.setFont(font)
        self.outputlabel.setObjectName("outputlabel")
        self.outputabundances2.addWidget(self.outputlabel, 3, 1, 1, 2)
        self.gridLayout.addWidget(self.outputabundances, 4, 1, 1, 1)
        MEAFS.setCentralWidget(self.centralwidget)
        self.menubar = QtWidgets.QMenuBar(parent=MEAFS)
        self.menubar.setGeometry(QtCore.QRect(0, 0, 1203, 30))
        self.menubar.setObjectName("menubar")
        self.menuFile = QtWidgets.QMenu(parent=self.menubar)
        self.menuFile.setObjectName("menuFile")
        self.menuView = QtWidgets.QMenu(parent=self.menubar)
        self.menuView.setObjectName("menuView")
        self.menuEdit = QtWidgets.QMenu(parent=self.menubar)
        self.menuEdit.setObjectName("menuEdit")
        MEAFS.setMenuBar(self.menubar)
        self.statusbar = QtWidgets.QStatusBar(parent=MEAFS)
        self.statusbar.setObjectName("statusbar")
        MEAFS.setStatusBar(self.statusbar)
        self.open = QtGui.QAction(parent=MEAFS)
        self.open.setObjectName("open")
        self.save = QtGui.QAction(parent=MEAFS)
        self.save.setObjectName("save")
        self.openabundances = QtGui.QAction(parent=MEAFS)
        self.openabundances.setObjectName("openabundances")
        self.new_2 = QtGui.QAction(parent=MEAFS)
        self.new_2.setCheckable(False)
        self.new_2.setObjectName("new_2")
        self.saveas = QtGui.QAction(parent=MEAFS)
        self.saveas.setObjectName("saveas")
        self.quit = QtGui.QAction(parent=MEAFS)
        self.quit.setObjectName("quit")
        self.fullspec = QtGui.QAction(parent=MEAFS)
        self.fullspec.setObjectName("fullspec")
        self.fitpar = QtGui.QAction(parent=MEAFS)
        self.fitpar.setObjectName("fitpar")
        self.autosave = QtGui.QAction(parent=MEAFS)
        self.autosave.setCheckable(True)
        self.autosave.setChecked(True)
        self.autosave.setObjectName("autosave")
        self.checkcontinuumplot = QtGui.QAction(parent=MEAFS)
        self.checkcontinuumplot.setObjectName("checkcontinuumplot")
        self.erasecontinuumplot = QtGui.QAction(parent=MEAFS)
        self.erasecontinuumplot.setObjectName("erasecontinuumplot")
        self.errorguireset = QtGui.QAction(parent=MEAFS)
        self.errorguireset.setObjectName("errorguireset")
        self.clearfinalplotsscale = QtGui.QAction(parent=MEAFS)
        self.clearfinalplotsscale.setObjectName("clearfinalplotsscale")
        self.menuFile.addAction(self.new_2)
        self.menuFile.addAction(self.open)
        self.menuFile.addSeparator()
        self.menuFile.addAction(self.openabundances)
        self.menuFile.addSeparator()
        self.menuFile.addAction(self.save)
        self.menuFile.addAction(self.saveas)
        self.menuFile.addAction(self.autosave)
        self.menuFile.addSeparator()
        self.menuFile.addAction(self.quit)
        self.menuView.addAction(self.fullspec)
        self.menuView.addAction(self.checkcontinuumplot)
        self.menuView.addAction(self.erasecontinuumplot)
        self.menuView.addAction(self.clearfinalplotsscale)
        self.menuEdit.addAction(self.fitpar)
        self.menuEdit.addAction(self.errorguireset)
        self.menubar.addAction(self.menuFile.menuAction())
        self.menubar.addAction(self.menuEdit.menuAction())
        self.menubar.addAction(self.menuView.menuAction())

        self.retranslateUi(MEAFS)
        self.methodsdatafittab.setCurrentIndex(2)
        self.methodstab.setCurrentIndex(0)
        self.tabplotshels.setCurrentIndex(0)
        self.plotstab.setCurrentIndex(2)
        self.histplottab.setCurrentIndex(0)
        QtCore.QMetaObject.connectSlotsByName(MEAFS)

    def retranslateUi(self, MEAFS):
        _translate = QtCore.QCoreApplication.translate
        MEAFS.setWindowTitle(_translate("MEAFS", "MEAFS"))
        item = self.dataloadtable.horizontalHeaderItem(1)
        item.setText(_translate("MEAFS", "Data"))
        self.delimitertype.setItemText(0, _translate("MEAFS", "Delimiter Type"))
        self.delimitertype.setItemText(1, _translate("MEAFS", "Comma"))
        self.delimitertype.setItemText(2, _translate("MEAFS", "Tab"))
        self.loaddata.setText(_translate("MEAFS", "Load Data"))
        self.methodsdatafittab.setTabText(self.methodsdatafittab.indexOf(self.datatab), _translate("MEAFS", "Data"))
        self.methodconfiglabel.setText(_translate("MEAFS", "Method Configuration"))
        self.methodbox.setItemText(0, _translate("MEAFS", "Equivalent Width"))
        self.methodbox.setItemText(1, _translate("MEAFS", "TurboSpectrum"))
        self.methodbox.setItemText(2, _translate("MEAFS", "PFANT"))
        self.methodbox.setItemText(3, _translate("MEAFS", "Synthe"))
        self.convinitguesslabel.setText(_translate("MEAFS", "Convolution"))
        self.deepthinitguesslabel.setText(_translate("MEAFS", "Deepth"))
        self.ewfuncbox.setItemText(0, _translate("MEAFS", "Gaussian"))
        self.ewfuncbox.setItemText(1, _translate("MEAFS", "Lorentzian"))
        self.ewfuncbox.setItemText(2, _translate("MEAFS", "Voigt"))
        self.initguesslabel.setText(_translate("MEAFS", "Inital Guess"))
        self.ewfunclabel.setText(_translate("MEAFS", "Profile to Fit"))
        self.methodstab.setTabText(self.methodstab.indexOf(self.eqwidthtab), _translate("MEAFS", "Eq. Width"))
        self.turbospectrumconfigbrowse.setText(_translate("MEAFS", "Browse"))
        self.turbospectrumoutputbrowse.setText(_translate("MEAFS", "Browse"))
        self.turbospectrumoutputlabel.setText(_translate("MEAFS", "TurboSpectrum Output Location"))
        self.turbospectrumconfiglabel.setText(_translate("MEAFS", "TurboSpectrum Configuration File Location"))
        self.methodstab.setTabText(self.methodstab.indexOf(self.turbspectrumtab), _translate("MEAFS", "TurboSpectrum"))
        self.label_3.setText(_translate("MEAFS", "Not Implemented"))
        self.methodstab.setTabText(self.methodstab.indexOf(self.pfanttab), _translate("MEAFS", "PFANT"))
        self.label_4.setText(_translate("MEAFS", "Not Implemented"))
        self.methodstab.setTabText(self.methodstab.indexOf(self.synthetab), _translate("MEAFS", "Synthe"))
        self.methodsdatafittab.setTabText(self.methodsdatafittab.indexOf(self.majormethodstab), _translate("MEAFS", "Method"))
        self.manualfitbutton.setToolTip(_translate("MEAFS", "This Option saves automatically."))
        self.manualfitbutton.setText(_translate("MEAFS", "Manual Fit"))
        self.currentvaluesplotbutton.setText(_translate("MEAFS", "Plot Current Values"))
        self.currentvaluessavebutton.setText(_translate("MEAFS", "Save Current Values"))
        self.linedeflabel.setText(_translate("MEAFS", "Current"))
        self.convolutionlabel.setText(_translate("MEAFS", "Convolution"))
        self.continuumlabel.setText(_translate("MEAFS", "Continuum"))
        self.abundancelabel.setText(_translate("MEAFS", "Abundance"))
        self.lambshiflabel.setText(_translate("MEAFS", "Lamb Shift"))
        self.methodsdatafittab.setTabText(self.methodsdatafittab.indexOf(self.fitparameterstab), _translate("MEAFS", "Fit Results"))
        self.linesplotlabel.setText(_translate("MEAFS", "Lines Plot"))
        self.histbinslabel.setToolTip(_translate("MEAFS", "Number of bins to create the histogram."))
        self.histbinslabel.setText(_translate("MEAFS", "N. of Bins"))
        self.boxplotcreate.setText(_translate("MEAFS", "Create"))
        self.boxplotlabel.setText(_translate("MEAFS", "Box Plot"))
        self.histbinsvalue.setToolTip(_translate("MEAFS", "Number of bins to create the histogram."))
        self.histplotlabel.setText(_translate("MEAFS", "Histogram"))
        self.linesplotall.setText(_translate("MEAFS", "All Lines"))
        self.abundshift.setToolTip(_translate("MEAFS", "Shift of the abundance to plot the blue shaded region in the element line."))
        self.abundshiftlabel.setToolTip(_translate("MEAFS", "Shift of the abundance to plot the blue shaded region in the element line."))
        self.abundshiftlabel.setText(_translate("MEAFS", "Abundance Shift"))
        self.histplotsingle.setText(_translate("MEAFS", "Single Element"))
        self.linesplotsingle.setText(_translate("MEAFS", "Single Line"))
        self.hitsplotall.setText(_translate("MEAFS", "All Elements"))
        self.methodsdatafittab.setTabText(self.methodsdatafittab.indexOf(self.finalplotsruntab), _translate("MEAFS", "Final Plots"))
        self.tabplotshels.setTabText(self.tabplotshels.indexOf(self.tabplot), _translate("MEAFS", "Spectrum Plot"))
        self.linesplotimage.setToolTip(_translate("MEAFS", "Demonstration only. Check the generated file for a better visualization."))
        self.linesplotimage.setText(_translate("MEAFS", "No data."))
        self.plotstab.setTabText(self.plotstab.indexOf(self.linesplottab), _translate("MEAFS", "Lines Plot"))
        self.boxplotimage.setToolTip(_translate("MEAFS", "Demonstration only. Check the generated file for a better visualization."))
        self.boxplotimage.setText(_translate("MEAFS", "No data."))
        self.plotstab.setTabText(self.plotstab.indexOf(self.boxplottab), _translate("MEAFS", "Box Plot"))
        self.abundhistplotimage.setToolTip(_translate("MEAFS", "Demonstration only. Check the generated file for a better visualization."))
        self.abundhistplotimage.setText(_translate("MEAFS", "No data."))
        self.histplottab.setTabText(self.histplottab.indexOf(self.abundhisttab), _translate("MEAFS", "Abundance"))
        self.differhistplotimage.setToolTip(_translate("MEAFS", "Demonstration only. Check the generated file for a better visualization."))
        self.differhistplotimage.setText(_translate("MEAFS", "No data."))
        self.histplottab.setTabText(self.histplottab.indexOf(self.differhisttab), _translate("MEAFS", "Difference"))
        self.plotstab.setTabText(self.plotstab.indexOf(self.histplottabs), _translate("MEAFS", "Histogram"))
        self.tabplotshels.setTabText(self.tabplotshels.indexOf(self.tabfinalplots), _translate("MEAFS", "Final Plots"))
        self.tabplotshels.setTabText(self.tabplotshels.indexOf(self.tabjupyshell), _translate("MEAFS", "Jupyter Shell"))
        self.clearstdbutton.setText(_translate("MEAFS", "Clear"))
        self.tabplotshels.setTabText(self.tabplotshels.indexOf(self.tabstoutstderr), _translate("MEAFS", "Stdout/Stderr"))
        self.linelistlabel.setText(_translate("MEAFS", "Linelist Options"))
        self.linelistload.setText(_translate("MEAFS", "Load"))
        self.linelistbrowse.setText(_translate("MEAFS", "Browse"))
        self.linelistlabel2.setText(_translate("MEAFS", "Linelist Location:"))
        self.linelistcheckbox.setText(_translate("MEAFS", "Select All"))
        self.linelistaddline.setText(_translate("MEAFS", "Add Line"))
        self.linelistremoveline.setText(_translate("MEAFS", "Remove Line"))
        self.restart.setToolTip(_translate("MEAFS", "If there are lines saved in the results file, they will be erased."))
        self.restart.setText(_translate("MEAFS", "Restart"))
        item = self.linelistcontent.horizontalHeaderItem(0)
        item.setText(_translate("MEAFS", " Element "))
        item = self.linelistcontent.horizontalHeaderItem(1)
        item.setText(_translate("MEAFS", " Wavelength "))
        item = self.linelistcontent.horizontalHeaderItem(2)
        item.setText(_translate("MEAFS", " Analyze "))
        self.referlabel.setText(_translate("MEAFS", "Abundance Reference"))
        self.referlabel2.setText(_translate("MEAFS", "Abund. Ref. Location:"))
        self.referaddline.setText(_translate("MEAFS", "Add Line"))
        self.referremoveline.setText(_translate("MEAFS", "Remove Line"))
        self.referbrowse.setText(_translate("MEAFS", "Browse"))
        self.referload.setText(_translate("MEAFS", "Load"))
        item = self.refercontent.horizontalHeaderItem(0)
        item.setText(_translate("MEAFS", "Element"))
        item = self.refercontent.horizontalHeaderItem(1)
        item.setText(_translate("MEAFS", "Abundance"))
        self.outputbrowse.setText(_translate("MEAFS", "Browse"))
        self.abundanceslabel.setText(_translate("MEAFS", "Abundances"))
        item = self.abundancetable.horizontalHeaderItem(0)
        item.setText(_translate("MEAFS", "Element"))
        item = self.abundancetable.horizontalHeaderItem(1)
        item.setText(_translate("MEAFS", "Wavelength"))
        self.stop.setText(_translate("MEAFS", "Stop"))
        self.progresslabel.setText(_translate("MEAFS", "Progress:"))
        self.run.setText(_translate("MEAFS", "Run"))
        self.progressvalue.setText(_translate("MEAFS", "0/0"))
        self.outputlabel.setText(_translate("MEAFS", "Output Location:"))
        self.menuFile.setTitle(_translate("MEAFS", "File"))
        self.menuView.setTitle(_translate("MEAFS", "View"))
        self.menuEdit.setTitle(_translate("MEAFS", "Edit"))
        self.open.setText(_translate("MEAFS", "&Open..."))
        self.save.setText(_translate("MEAFS", "&Save"))
        self.openabundances.setText(_translate("MEAFS", "O&pen Abundances..."))
        self.new_2.setText(_translate("MEAFS", "&New..."))
        self.saveas.setText(_translate("MEAFS", "Sa&ve as..."))
        self.quit.setText(_translate("MEAFS", "&Quit"))
        self.fullspec.setText(_translate("MEAFS", "&Change Plot to full spectrum range (Ctrl+Shift+V)"))
        self.fullspec.setToolTip(_translate("MEAFS", "Change Plot to full spectrum range (CRTL+V)"))
        self.fitpar.setText(_translate("MEAFS", "&Fit Parameters"))
        self.autosave.setText(_translate("MEAFS", "&Auto Save"))
        self.autosave.setToolTip(_translate("MEAFS", "TESTE"))
        self.checkcontinuumplot.setText(_translate("MEAFS", "C&heck Continuum"))
        self.erasecontinuumplot.setText(_translate("MEAFS", "&Erase Continuum"))
        self.errorguireset.setText(_translate("MEAFS", "&Error GUI Reset"))
        self.clearfinalplotsscale.setText(_translate("MEAFS", "Clear Final Plots Scale"))


if __name__ == "__main__":
    import sys
    app = QtWidgets.QApplication(sys.argv)
    MEAFS = QtWidgets.QMainWindow()
    ui = Ui_MEAFS()
    ui.setupUi(MEAFS)
    MEAFS.show()
    sys.exit(app.exec())



"""
| MEAFS Fit Functions
| Matheus J. Castro

| Fit functions that are used, like the C library, the :math:`\\chi^2` calculation, bisection and more.
"""

from astropy.convolution import Gaussian1DKernel, convolve
from pathlib import Path
import numpy as np
import subprocess
import ctypes
import os


def c_init(c_name):
    """
    Function to initialize the C shared library

    :param c_name: name of the shared C library.
    :return: the initialized object.
    """

    c_library = ctypes.CDLL("{}".format(c_name))

    # Defining functions argument types
    c_library.bisec.argtypes = [ctypes.POINTER(ctypes.c_float), ctypes.c_int, ctypes.c_float]
    c_library.bisec.restype = ctypes.c_int

    c_library.chi2.argtypes = [ctypes.POINTER(ctypes.c_float), ctypes.POINTER(ctypes.c_float), ctypes.c_int,
                               ctypes.POINTER(ctypes.c_float), ctypes.POINTER(ctypes.c_float), ctypes.c_int]
    c_library.chi2.restype = ctypes.c_float

    return c_library


def chi2(spec1, spec2):
    """
    Function to find the :math:`\\chi^2` of two arrays using the C library.

    :param spec1: first array.
    :param spec2: seccond array.
    :return: the :math:`\\chi^2`.
    """

    global c_lib

    # Using C library
    spec1x = spec1[0].tolist()
    spec1y = spec1[1].tolist()

    # noinspection PyCallingNonCallable,PyTypeChecker
    spec1x = (ctypes.c_float * len(spec1x))(*spec1x)
    # noinspection PyCallingNonCallable,PyTypeChecker
    spec1y = (ctypes.c_float * len(spec1y))(*spec1y)

    spec2x = spec2[0].tolist()
    spec2y = spec2[1].tolist()

    # noinspection PyCallingNonCallable,PyTypeChecker
    spec2x = (ctypes.c_float * len(spec2x))(*spec2x)
    # noinspection PyCallingNonCallable,PyTypeChecker
    spec2y = (ctypes.c_float * len(spec2y))(*spec2y)

    return c_lib.chi2(spec1x, spec1y, len(spec1x), spec2x, spec2y, len(spec2x))

    # Using Python (slower)
    # sp1_interpol = np.array([])
    # sp2_interpol = np.array([])
    #
    # for i in spec2.iloc():
    #     pos = bisec(spec1, i[0])
    #     if pos != -1:
    #         x1 = spec1.iloc[pos][0]
    #         x2 = spec1.iloc[pos + 1][0]
    #         y1 = spec1.iloc[pos][1]
    #         y2 = spec1.iloc[pos + 1][1]
    #         x = i[0]
    #         y = y1 + (y2 - y1) / (x2 - x1) * (x - x1)
    #
    #         sp1_interpol = np.append(sp1_interpol, y)
    #         sp2_interpol = np.append(sp2_interpol, i[1])
    #
    # return sum((sp1_interpol - sp2_interpol)**2 / sp2_interpol)


def bisec(spec, lamb):
    """
    Function to apply the bisection script to find a number position in an array.

    :param spec: array to analyse.
    :param lamb: value to look for.
    :return: the position.
    """

    # Uncomment the script bellow to use the C library (slower)
    # global c_lib
    # arr = spec[0].tolist()
    # arr_c = (ctypes.c_float * len(arr))(*arr)
    #
    # index = c_lib.bisec(arr_c, len(arr), lamb)
    #
    # return index

    # Bisection Script
    if lamb < spec.iloc[0][0]:
        return -1
    elif lamb > spec.iloc[-1][0]:
        return -1
    else:
        gap = [0, len(spec) - 1]

        while True:
            new_gap = gap[0] + (gap[1] - gap[0]) // 2

            if lamb < spec.iloc[new_gap][0]:
                gap[1] = new_gap
            else:
                gap[0] = new_gap

            if gap[0] + 1 == gap[1]:
                return gap[0]


def cut_spec(spc, lamb, cut_val=1.):
    """
    Function to restrict the array in the desired range

    :param spc: the array.
    :param lamb: the central position.
    :param cut_val: the range to be restricted.
    :return: the restricted array.
    """

    val0 = bisec(spc, lamb - cut_val)
    val1 = bisec(spc, lamb + cut_val)+1

    return spc.iloc[val0:val1]


def spec_operations(spec, lamb_desloc=0., continuum=1., convol=0.):
    """
    Function to apply lambda shift, continuum fit and convolution to the spectrum

    :param spec: the spectrum to be changed.
    :param lamb_desloc: the shift in wavelength to be applied.
    :param continuum: the continuum to be applied.
    :param convol: the convolution to be applied.
    :return: the modified spectrum.
    """

    spec[0] = spec[0] + lamb_desloc
    spec[1] = spec[1] * continuum

    if convol != 0:
        try:
            # Apply the convolution using the Gaussian1DKernel function
            g = Gaussian1DKernel(stddev=convol)
            spec[1] = convolve(spec[1], g)  # convolution between the spectrum and the function above
        except ValueError:
            pass

    return spec


def fit_continuum(spec, contpars=None, iterac=1000):
    """
    Fit the overall continuum in the entire spectrum.

    :param spec: spectrum data.
    :param contpars: the calibration values of the method.
    :param iterac: maximum number of iterations.
    :return: the mean and the standard deviation (continuum and errors).
    """

    if contpars is None:
        alpha = .5
        eps = 20
    else:
        alpha = contpars[0]
        eps = contpars[1]

    eps = np.float16(10)**(-np.float16(eps))

    new_spec = spec.iloc[:, 1].values.tolist()
    median = np.median(new_spec)
    std = np.std(new_spec)
    std_old = std
    count = 0

    while True:
        for i in reversed(range(len(new_spec))):
            value = new_spec[i]
            if value > median + alpha * std or value < median - alpha * std:
                del new_spec[i]

        median = np.median(new_spec)
        std = np.std(new_spec)

        count += 1
        if count >= iterac:
            print("Maximum Iterations for Continuum fit reached. iter_max = ", iterac)
            break
        elif (std_old - std) / std <= eps:
            break
        std_old = std

    median = np.median(new_spec)
    std = np.std(new_spec)

    return median, std


# Compile C files
if not os.path.isfile(Path(os.path.dirname(__file__)).joinpath("bisec_interpol.so")):
    currdir = os.getcwd()
    os.chdir(Path(os.path.dirname(__file__)))
    subprocess.run("gcc -Wall -pedantic -O3 bisec_interpol.c -o bisec_interpol.so -shared", shell=True)
    os.chdir(currdir)

c_lib = c_init(Path(os.path.dirname(__file__)).joinpath("bisec_interpol.so"))  # initialize the C library
